# coding: utf-8

"""
The Sublocus class is the first to be invoked during the Mikado pick analysis.
Each of these containers holds transcripts which either are monoexonic and overlapping,
or multiexonic and with at least one intron in common.
"""

import itertools
import importlib
import logging
from ..transcripts.transcript import Transcript
from .abstractlocus import Abstractlocus
from .excluded import Excluded
from .monosublocus import Monosublocus
from ..parsers.GFF import GffLine
from ..scales import c_compare
from ..utilities.log_utils import create_null_logger
from ..utilities import overlap


# pylint: disable=too-many-instance-attributes
class Sublocus(Abstractlocus):
    """
    The sublocus class is created either by the superlocus class during
    the subloci definition, or directly using a G(T|F)line-like object.
    It is used to define the monosubloci.
    """

    __name__ = "sublocus"
    available_metrics = Transcript.get_available_metrics()

    # ############### Class special methods ##############

    def __init__(self,
                 transcript_instance=None,
                 configuration=None,
                 logger=None,
                 verified_introns=None,
                 **kwargs):

        """
        :param transcript_instance: an instance which describes a genomic interval
        :type transcript_instance: Transcript | Abstractlocus

        :param configuration: a configuration dictionary
        :type configuration: (MikadoConfiguration|DaijinConfiguration)

        :param logger: a logger instance from the logging library
        :type logger: logging.Logger | None

        This class takes as input a "span" feature - e.g. a GffLine or a transcript_instance.
        The span instance should therefore have such attributes as
        chrom, strand, start, end, attributes.
        """

        self.counter = 0  # simple tag for avoiding collisions in the GFF output
        self.__splitted = False
        Abstractlocus.__init__(self, verified_introns=verified_introns,
                               configuration=configuration,
                               logger=logger,
                               **kwargs)
        self.feature = self.__name__
        self.logger.debug("Verified introns for %s: %s", self.id, verified_introns)
        self.fixed_size = False
        self.source = self.configuration.pick.output_format.source

        self.excluded = None
        self._not_passing = set()
        self.splitted = False
        # Flag to indicate that we have not calculated the metrics for the transcripts
        # Flag to indicate that we have not calculated the scores for the transcripts
        setattr(self, "monoexonic", getattr(transcript_instance, "monoexonic", None))
        if transcript_instance is not None:
            if transcript_instance.__name__ == "transcript":
                transcript_instance.finalize()
            if configuration is None and transcript_instance.configuration is not None:
                configuration = transcript_instance.configuration
            if transcript_instance and transcript_instance.feature == "sublocus":
                self.fixed_size = True

        self.configuration = configuration

        # This part is necessary to import modules
        if hasattr(self.configuration, "modules"):
            for mod in self.configuration.modules:
                globals()[mod] = importlib.import_module(mod)

        if isinstance(transcript_instance, Transcript):
            self.add_transcript_to_locus(transcript_instance)
            self.attributes = transcript_instance.attributes.copy()
            self.parent = transcript_instance.parent
        else:
            self.parent = getattr(transcript_instance, "parent", None)
            self.chrom = getattr(transcript_instance, "chrom", None)
            self.start = getattr(transcript_instance, "start", None)
            self.end = getattr(transcript_instance, "end", None)
            self.strand = getattr(transcript_instance, "strand", None)
            self.attributes = getattr(transcript_instance, "attributes", dict())

        self.monosubloci = []
        self.logger.debug("Initialized {0}".format(self.id))
        self.metric_lines_store = []  # This list will contain the lines to be printed in the metrics file
        self.excluded = Excluded(configuration=configuration)
        self.scores = dict()

    # pylint: disable=arguments-differ
    def __str__(self, print_cds=True):

        lines = []

        self_line = GffLine('')
        for attr in ["chrom", 'feature', 'source', 'start', 'end', 'strand']:
            setattr(self_line, attr, getattr(self, attr))
        self_line.phase, self_line.score = None, None
        self_line.id = "{0}_{1}".format(self.source, self.id)
        self_line.name = self.name
        self_line.parent = self.parent
        self_line.attributes["multiexonic"] = (not self.monoexonic)
        lines.append(str(self_line))

        for tid in sorted(self.transcripts, key=lambda ttid: self.transcripts[ttid]):
            self.transcripts[tid].source = self.source
            self.transcripts[tid].parent = self_line.id
            lines.append(self.transcripts[tid].format(
                "gff3",
                all_orfs=self.configuration.pick.output_format.report_all_orfs,
                with_cds=print_cds).rstrip())

        return "\n".join(lines)
    # pylint: enable=arguments-differ

    # ########## Class instance methods #####################

    def as_dict(self):
        state = super().as_dict()
        state["monosubloci"] = [_.as_dict() for _ in self.monosubloci]
        state["excluded"] = self.excluded.as_dict()
        return state

    def load_dict(self, state, load_transcripts=True, load_configuration=True):
        super().load_dict(state, load_transcripts=load_transcripts, load_configuration=load_configuration)
        self.monosubloci = []
        for stat in state["monosubloci"]:
            s = Monosublocus()
            s.load_dict(stat)
            self.monosubloci.append(s)
        self.excluded = Excluded(configuration=self.configuration)
        self.excluded.load_dict(state["excluded"])

    def add_transcript_to_locus(self, transcript: Transcript, check_in_locus=True, **kwargs):

        """
        :param transcript: the transcript which might be putatively added to the Locus.
        :type transcript: Transcript

        :param check_in_locus: whether to check that the transcript is compatible. Boolean, default True.
        :type check_in_locus: bool

        :param kwargs: eventual keyword arguments are ignored.

        This is an override of the original method, as at the sublocus stage
        we need to accomplish a couple of things more:
        - check that transcripts added to the sublocus are either all monoexonic or all multiexonic
        - change the id of the transcripts to the new ID
        """

        _ = kwargs  # Ignore any keyword arguments

        if len(self.transcripts) > 0:
            self.logger.debug("Adding %s to %s",
                              transcript.id, self.id)
        else:
            self.logger.debug("Initializing sublocus with %s at %s%s:%d-%d",
                              transcript.id,
                              transcript.chrom,
                              transcript.strand,
                              transcript.start,
                              transcript.end)

        if transcript is None:
            return

        cds_only = self.configuration.pick.clustering.cds_only

        monoexonic = self._is_transcript_monoexonic(transcript)

        if self.initialized is False:
            self.monoexonic = monoexonic
            self.logger.debug("Locus %s is %s because of %s that is monoexonic: %s",
                              self.id, self.monoexonic, transcript.id, monoexonic)

        elif self.monoexonic != monoexonic:
            raise ValueError("""Sublocus and transcript are not compatible!
                            {0}\t{1}\t{2}\t{3}\t
                            Locus monoexonic: {4}; CDS only: {5}
                            Wrong transcript:
                            {6}
                            Others:
                            {7}
                            """.format(self.chrom, self.start, self.end, self.strand, self.monoexonic, cds_only,
                                       transcript.format("bed12") + "\t{}".format(
                                           self._is_transcript_monoexonic(transcript)),
                                       "\n".join([_.format("bed12") + "\t{}".format(
                                           self._is_transcript_monoexonic(_)) for _ in self.transcripts.values()])))
        super().add_transcript_to_locus(transcript, check_in_locus=check_in_locus)
        # add the verified introns from the outside
        self.logger.debug("Added %s to %s", transcript.id, self.id)
        # Update the id

    def define_monosubloci(self, purge=False, check_requirements=True):
        """
        :param excluded: the excluded Locus to which transcripts from purged loci will be added to
        :type excluded: None
        :type excluded: Mikado.loci_objects.excluded.Excluded

        :param purge: a flag which indicates whether loci whose
        best transcript has a score of 0 should be excluded (True) or retained (False)
        :type purge: bool

        :param check_requirements: Boolean flag, true by default. If set to false transcripts requirements will not
        be evaluated.
        :type check_requirements: bool

        This function retrieves the best non-overlapping transcripts inside
        the sublocus, according to the score calculated by
        calculate_scores (explicitly called inside the method).
        The "excluded" keyword must contain either None or
        a MonosublocusHolder object. It is used to contain
        transcripts that must be excluded from the Locus due to unmet requirements.
        """

        self.monosubloci = []
        # self.excluded = excluded
        self.logger.debug("Launching calculate scores for {0}".format(self.id))
        self.filter_and_calculate_scores(check_requirements=check_requirements)

        if len(self._excluded_transcripts) > 0 and self.purge:
            excluded_tids = list(self._excluded_transcripts.keys())
            self.excluded.add_transcripts_to_locus(
                self._excluded_transcripts.values(), unsafe=True
            )
            self.remove_transcripts_from_locus(set.intersection(set(excluded_tids),
                                                                set(self.transcripts.keys())))
            [self._excluded_transcripts.pop(key, None) for key in excluded_tids]

        self.logger.debug("Defining monosubloci for {0}".format(self.id))

        transcript_graph = self.define_graph(self.transcripts,
                                             inters=self.is_intersecting,
                                             logger=self.logger)

        while len(transcript_graph) > 0:
            communities = self.find_communities(transcript_graph)
            self.logger.debug("Communities: {0}".format(communities))
            to_remove = set()
            for msbl in communities:
                msbl = {x: self.transcripts[x] for x in msbl}
                selected_tid = self.choose_best(msbl)
                selected_transcript = self.transcripts[selected_tid]
                to_remove.add(selected_tid)
                self.logger.debug("Selected: %s (score: %f)",
                                  selected_tid, selected_transcript.score)
                self.logger.debug("Removing as intersecting {0}: {1}".format(
                            selected_tid,
                            ",".join(set(transcript_graph.neighbors(selected_tid)))
                        ))
                to_remove.update(set(transcript_graph.neighbors(selected_tid)))
                if purge is False or selected_transcript.score > 0:
                    new_locus = Monosublocus(selected_transcript,
                                             logger=self.logger,
                                             configuration=self.configuration,
                                             use_transcript_scores=self._use_transcript_scores)
                    new_locus.configuration = self.configuration
                    self.monosubloci.append(new_locus)
            if not to_remove:
                message = "No transcripts to remove from the pool for {0}\n".format(self.id)
                message += "Transcripts remaining: {0}".format(communities)
                exc = ValueError(message)
                self.logger.exception(exc)
                raise exc
            self.logger.debug("Removing %d from transcripts for %s",
                              len(to_remove), self.id)
            transcript_graph.remove_nodes_from(to_remove)
        self.logger.debug("Defined monosubloci for %s", self.id)
        self.splitted = True
        self.logger.debug("Defined monosubloci for %s", self.id)
        return self.excluded

    def load_scores(self, scores):
        """
        :param scores: an external dictionary with scores
        :type scores: dict

        Simple mock function to load scores for the transcripts from an external dictionary.
        """

        for tid in self.transcripts:
            if tid in scores:
                self.transcripts[tid].score = scores[tid]
            else:
                self.transcripts[tid].score = 0

    def prepare_metrics(self):

        """This method prepares the dictionary "rows"
        that will be given to a csv.DictWriter class."""

        for row in Abstractlocus.print_metrics(self):
            yield row
        if self.excluded is not None:
            for row in self.excluded.print_metrics():
                yield row
        return

    def print_metrics(self):

        """
        THis method yields the dictionary "rows" that will be given to a csv.DictWriter class.
        :return:
        """

        # self.filter_and_calculate_scores()
        self.get_metrics()
        self.metric_lines_store = [_ for _ in self.prepare_metrics()]
        for row in self.metric_lines_store:
            yield row

    def print_scores(self):
        """This method yields dictionary rows that are given to a csv.DictWriter class."""
        self.filter_and_calculate_scores()
        score_keys = sorted(list(self.configuration.scoring.scoring.keys()) + ["source_score"])
        keys = ["tid", "alias", "parent", "score"] + sorted(score_keys)

        for tid in self.scores:
            row = dict().fromkeys(keys)
            row["tid"] = tid
            row["alias"] = self.transcripts[tid].alias
            row["parent"] = self.id
            row["score"] = round(self.scores[tid]["score"], 2)
            for key in score_keys:
                assert key in self.scores[tid] and self.scores[tid][key] != "NA" and self.scores[tid][key] is not None, (key, self.scores[tid].keys())
                row[key] = round(self.scores[tid][key], 2)

            if tid not in self._not_passing:
                score_sum = sum(row[key] for key in score_keys)

                if round(score_sum, 2) != round(self.scores[tid]["score"], 2):
                    try:
                        scores = dict(_ for _ in self.scores[tid].items())
                    except KeyError as exc:
                        raise KeyError((exc, row.keys()))
                    recalc = dict(_ for _ in row.items() if _[0] not in ["tid", "parent", "score"])
                    error = AssertionError("Tid: {}; Sum: {}; Calculated: {}\nScores: {}\nRecalculated: {}".format(
                        tid,
                        score_sum,
                        self.transcripts[tid].score,
                        scores,
                        recalc
                    ))
                    self.logger.exception(error)
                    raise error

            yield row

    # ############## Class methods ################

    # Class specific implementation differs from abstract blueprint
    # pylint: disable=arguments-differ
    @classmethod
    def is_intersecting(cls,
                        transcript,
                        other,
                        cds_only=False,
                        logger=None,
                        min_cdna_overlap=0.2,
                        min_cds_overlap=0.2,
                        simple_overlap_for_monoexonic=True) -> bool:
        """
        Implementation of the is_intersecting method. Now that we are comparing transcripts that
        by definition span multiple subloci, we have to be less strict in our definition of what
        counts as an intersection.
        Criteria:
        - the cDNA and CDS overlap is over a user-specified threshold
        OR
        - either transcript is monoexonic and simple_overlap_for_monoexonic is True, if there is exonic overlap
        OR
        - one intron of either transcript is completely contained within an exon of the other.

        The user can specify whether she prefers to consider the whole transcript (default) or whether to consider
        instead the **selected ORF** of the transcripts for the comparison. Please note that intersection in secondary
        ORFs will not be valid under this scenario.

         :param transcript
         :type transcript; Transcript

         :param other:
         :type other: Transcript

         :param cds_only: boolean flag. If set to True, only the CDS component of the transcripts will be
         considered to determine whether they are intersecting or not.
         :type cds_only: bool

        :param min_cdna_overlap: float. This is the minimum cDNA overlap for two transcripts to be considered as intersecting,
         even when all other conditions fail.
        :type min_cdna_overlap: float

        :param min_cds_overlap: float. This is the minimum CDS overlap for two transcripts to be considered as intersecting,
         even when all other conditions fail.
        :type min_cds_overlap: float

        :param simple_overlap_for_monoexonic: boolean flag. If set to true, any overlap for monoexonic transcripts
        will be enough to trigger incorporation in the locus.
        :type simple_overlap_for_monoexonic: bool

        :param logger: either None or a logger instance. If None, a null logger will be created.

         :rtype : bool
        """

        if logger is None or not isinstance(logger, logging.Logger):
            logger = create_null_logger("MSH")
        transcript.finalize()
        other.finalize()

        logger.debug("Comparing %s vs. %s", transcript.id, other.id)

        if transcript.id == other.id or transcript.strand != other.strand:
            logger.debug("Cannot intersect with itself (%s vs %s) or a transcript on the other strand (%s and %s)",
                         transcript.id, other.id, transcript.strand, other.strand)
            return False

        if cds_only is True and transcript.is_coding and other.is_coding:
            logger.debug("Consider only the CDS: %s", cds_only)
            if overlap((transcript._selected_orf_transcript.start, transcript._selected_orf_transcript.end),
                       (other._selected_orf_transcript.start, other._selected_orf_transcript.end)) <= 0:
                intersecting, reason = False, "No genomic overlap between {} and {}".format(
                    transcript.id, other.id
                )
            else:
                intersecting, reason = cls._transcripts_are_intersecting(
                    transcript._selected_orf_transcript,
                    other._selected_orf_transcript,
                    min_cdna_overlap=min_cdna_overlap,
                    min_cds_overlap=min_cds_overlap,
                    simple_overlap_for_monoexonic=simple_overlap_for_monoexonic)
        else:
            if overlap((transcript.start, transcript.end),
                       (other.start, other.end)) <= 0:
                intersecting, reason = False, "No genomic overlap between {} and {}".format(
                    transcript.id, other.id
                )
            else:
                intersecting, reason = cls._transcripts_are_intersecting(
                    transcript,
                    other,
                    min_cdna_overlap=min_cdna_overlap,
                    min_cds_overlap=min_cds_overlap,
                    simple_overlap_for_monoexonic=simple_overlap_for_monoexonic)

        logger.debug(reason)
        return intersecting

    @classmethod
    def _transcripts_are_intersecting(cls,
                                      transcript: Transcript,
                                      other: Transcript,
                                      min_cdna_overlap=0.2,
                                      min_cds_overlap=0.2,
                                      simple_overlap_for_monoexonic=True):
        """Private method which is called by is_intersecting. It decouples the determination of whether two transcripts
        intersect from the public interface of the method.
        :param transcript
         :type transcript; Transcript

         :param other:
         :type other: Transcript

        :param min_cdna_overlap: float. This is the minimum cDNA overlap for
        two transcripts to be considered as intersecting, even when all other conditions fail.
        :type min_cdna_overlap: float

        :param min_cds_overlap: float. This is the minimum CDS overlap for two transcripts to be considered as intersecting,
         even when all other conditions fail.
        :type min_cds_overlap: float

        :param simple_overlap_for_monoexonic: boolean flag. If set to true, any overlap for monoexonic transcripts
        will be enough to trigger incorporation in the locus.
        :type simple_overlap_for_monoexonic: bool

        :param is_internal_orf: boolean. Set to True if we are considering only the CDS for this run.
        :type is_internal_orf: bool
        """

        intersecting, reason = cls._simple_intersection_test(transcript, other, simple_overlap_for_monoexonic)
        if intersecting:
            return intersecting, reason

        comparison, _ = c_compare(other, transcript)
        if comparison.n_f1[0] == 0:
            reason = "No genomic overlap between {} and {}".format(transcript.id, other.id)
            intersecting = False
            return intersecting, reason
        elif comparison.j_f1[0] > 0 or comparison.ccode[0] == "h":
            reason = "{} and {} intersect; class code: {}".format(transcript.id, other.id, comparison.ccode[0])
            intersecting = True
        else:
            intersecting, reason = cls._evaluate_transcript_overlap(
                transcript, other,
                min_cdna_overlap=min_cdna_overlap,
                min_cds_overlap=min_cds_overlap,
                comparison=comparison,
                fixed_perspective=False)

        return intersecting, reason

    @classmethod
    def _simple_intersection_test(cls, transcript: Transcript, other:Transcript,
                                  simple_overlap_for_monoexonic: bool):
        """Private method to perform a fast check on transcript intersection.
        This method performs preliminary checks that do not require the full use of the
        transcript comparison method in Mikado.scales.
        :param transcript: first transcript to evaluate
        :param other: second transcript to evaluate
        :param simple_overlap_for_monoexonic: boolean flag.
        """

        intersecting, reason = False, None
        if any([transcript.monoexonic, other.monoexonic]) and simple_overlap_for_monoexonic:
            if all([transcript.monoexonic and other.monoexonic]):
                reason = f"{transcript.id} and {other.id} are monoexonic and on the same genomic region."
                intersecting = True
            elif transcript.monoexonic:
                exon = list(transcript.exons)[0]
                if any(other.segmenttree.find(*exon, value="exon", strict=False)):
                    reason = f"Simple overlap for monoexonic transcripts, for {transcript.id} and {other.id}"
                    intersecting = True
            elif other.monoexonic:
                exon = list(other.exons)[0]
                if any(transcript.segmenttree.find(*exon, value="exon", strict=False)):
                    reason = f"Simple overlap for monoexonic transcripts, for {transcript.id} and {other.id}"
                    intersecting = True
        elif not transcript.monoexonic and not other.monoexonic and set.intersection(transcript.splices,
                                                                                     other.splices):
            reason = f"{transcript.id} and {other.id} share at least one splice site."
            intersecting = True
        elif cls._intron_contained_in_exon(transcript, other) or cls._intron_contained_in_exon(other, transcript):
            reason = "Intronic containment within an exon for the comparison {} and {}; intersecting".format(
                transcript.id, other.id)
            intersecting = True

        return intersecting, reason

    @staticmethod
    def _intron_contained_in_exon(transcript: Transcript, other: Transcript) -> bool:

        """Mini-method to assess whether at least one intron of "transcript" is **completely** contained
        within an exon of "other"."""

        return any(
            transcript.segmenttree.find(*exon, strict=True, value="intron")
            for exon in other.exons
        )

    @property
    def splitted(self):
        """The splitted flag indicates whether a sublocus
        has already been processed to produce the necessary monosubloci.
        It must be set as a boolean flag (hence why it is coded as a property)

        :rtype bool
        """
        return self.__splitted

    @splitted.setter
    def splitted(self, verified):
        """
        :param verified: boolean flag to set.
        :type verified: bool
        """

        if not isinstance(verified, bool):
            raise TypeError()
        self.__splitted = verified

    @property
    def id(self):
        """
        :return: The name of the Locus.
        :rtype str
        """
        if self.monoexonic is True:
            addendum = "mono"
        else:
            addendum = "multi"
        if self.counter > 0:
            addendum = "{0}.{1}".format(addendum, self.counter)

        return "{0}.{1}".format(super().id, addendum)



#!/usr/bin/env python3
# coding: utf-8

"""Script to merge BED12 files *based on the thickStart/End features*.
Necessary for merging junction files such as those produced by TopHat."""

import Mikado.parsers.bed12
import sys
import collections
import operator
import argparse
import io
import multiprocessing


def serialise(filename, tophat=False):
    """Function to serialise the BED12. Useful for multiprocessing.

    :param filename: name of the file to analyse
    :type filename: io.TextIOWrapper | str

    :param tophat: flag, derived from Namespace
    :type tophat: bool

    :rtype : (dict, str)
    """

    juncs = collections.defaultdict(list)

    if type(filename) is str:
        bed = Mikado.parsers.bed12.Bed12Parser(open(filename))
    else:
        if type(filename) is not io.TextIOWrapper:
            raise TypeError("Invalid BED file: {0}".format(type(filename)))
        bed = Mikado.parsers.bed12.Bed12Parser(filename)

    header = ''
    for record in bed:
        if record.header is True:
            if header == '' and record._line.rstrip() != '':
                header = record._line.rstrip()
            continue
        if tophat is True:
            record.thickStart = record.start+record.blockSizes[0]-1
            record.thickEnd = record.start+record.blockStarts[1]-1

        key = (record.chrom, record.strand, record.thickStart, record.thickEnd)
        juncs[key].append(record)

    return juncs, header


def main():
    """
    Main function of the script.
    :return: None
    """

    parser = argparse.ArgumentParser("""Script to merge BED12 files *based on the thickStart/End features*.
    Necessary for merging junction files such as those produced by TopHat""")
    parser.add_argument("--delim", default=";", type=str,
                        help="Delimiter for merged names. Default: %(default)s")
    parser.add_argument("-t", "--threads", type=int, default=1,
                        help="Number of threads to use for multiprocessing. Default: %(default)s")
    parser.add_argument("--tophat", default=False, action="store_true",
                        help="""Flag. If set, tophat-like junction style is assumed. This means that junctions
                        are defined using the blockSizes rather than thickStart/End. The script will convert
                        the lines to this latter format. By default, the script assumes that the intron start/end
                        are defined using thickStart/End like in portcullis.
                        Mixed-type input files are not supported.""")
    parser.add_argument("-o", "--output", default=sys.stdout, type=argparse.FileType("w"),
                        help="Output file. Default: stdout")
    parser.add_argument("bed", type=str, nargs="+",
                        help="Input BED files. Use \"-\" for stdin.")
    args = parser.parse_args()

    if args.bed[0] == "-":
        args.bed = [Mikado.parsers.bed12.Bed12Parser(sys.stdin)]
        results = [serialise(sys.stdin, args.tophat)]
    else:
        # The process pool has size which is the minimum of
        # number of available cores
        # number of BED files
        # number of requested cores
        pool = multiprocessing.Pool(processes=min(len(args.bed),
                                                  args.threads,
                                                  multiprocessing.cpu_count()
                                                  ))
        results = pool.starmap(serialise, zip(args.bed, [args.tophat]*len(args.bed)))
        pool.close()
        pool.join()

    header = ''
    juncs = collections.defaultdict(list)

    for res in results:
        if header == '' and res[1] != '':
            header = res[1]
        else:
            pass
        for key in res[0]:
            juncs[key].extend(res[0][key])

    print(header, file=args.output)
    # Sort by chrom, then start, then end, finally strand
    for key in sorted(juncs.keys(), key=operator.itemgetter(0, 2, 3, 1)):
        newrecord = Mikado.parsers.bed12.BED12(None)
        newrecord.header = False
        newrecord.chrom = key[0]
        newrecord.start = min(x.start for x in juncs[key])
        newrecord.end = max(x.end for x in juncs[key])
        newrecord.strand = key[1]
        newrecord.name = args.delim.join(x.name for x in juncs[key])
        newrecord.score = sum(x.score if x.score is not None else 0 for x in juncs[key])
        newrecord.thick_start = key[2]+1
        newrecord.thick_end = key[3]
        newrecord.rgb = "255,0,0"
        newrecord.block_count = 2
        newrecord.block_sizes = [newrecord.thick_start-newrecord.start+1,
                                newrecord.end-newrecord.thick_end+1]
        newrecord.block_starts = [0, newrecord.thick_end - newrecord.start+1]
        print(newrecord, file=args.output)

if __name__ == "__main__":
    main()


__author__ = 'venturil'
__version__ = "0.8"



import unittest
from shutil import which

import pkg_resources
import os
import tempfile

import pytest
import sqlalchemy

from Mikado.serializers.blast_serializer import Hit
from Mikado.utilities import dbutils

from Mikado.serializers import blast_serializer
import sqlalchemy.orm
from Mikado import create_default_logger
from Mikado.configuration import MikadoConfiguration
from Mikado.subprograms.serialise import load_blast


class TestBlastParsing(unittest.TestCase):

    def setUp(self) -> None:
        self.master = pkg_resources.resource_filename("Mikado.tests", "blast_input_tests")
        self.queries = os.path.join(self.master, "queries.fasta")
        self.assertTrue(os.path.exists(self.queries))

    def run_loading(self, target, out_db, logger, conf):
        with self.assertLogs(logger, level="DEBUG") as cmo:
            try:
                blast_serializer.BlastSerializer(
                    target,
                    configuration=conf,
                    logger=logger)()
            except (ValueError, KeyError, TypeError) as exc:
                print(*cmo.output, sep="\n")
                raise exc
        self.assertTrue(os.path.exists(out_db), cmo.output)
        self.assertGreater(os.stat(out_db).st_size, 0)
        conf.db_settings.db = out_db
        engine = dbutils.connect(configuration=conf)
        Session = sqlalchemy.orm.sessionmaker()
        session = Session(bind=engine)
        self.assertTrue(session.query(Hit).count() > 0)

    @unittest.skip
    @pytest.mark.triage
    @unittest.skipUnless(which("blast_formatter") is not None, reason="NCBI BLAST+ not installed")
    def test_asn(self):
        # Currently DISABLED because the ASN specifications requires the database to be where indicated by the
        # relative path within the ASN. So for the time being this test is *not active*.

        asns = [os.path.join("blast", "asn", "blast.asn.gz"),
                os.path.join("blast_parse_seqids", "asn", "blast.asn.gz")]

        for folder in ["sanitised", "uniprot"]:
            conf = MikadoConfiguration()
            targets = os.path.join(self.master, folder, "uniprot.fasta")
            self.assertTrue(os.path.exists(targets))
            for asn in asns:
                asn = os.path.join(self.master, folder, asn)
                self.assertTrue(os.path.exists(asn))
                with tempfile.TemporaryDirectory() as out_folder:
                    conf.serialise.files.output_dir = out_folder
                    conf.serialise.files.blast_targets = [targets]
                    conf.serialise.files.transcripts = self.queries
                    conf.serialise.files.xml = [asn]
                    logger = create_default_logger(f"test_asn_{folder}")
                    out_db = os.path.join(out_folder, conf.db_settings.db)
                    conf.db_settings.db = out_db
                    self.run_loading(asn, out_db, logger, conf)

    @unittest.skipUnless(which("diamond") is not None, reason="DIAMOND not installed")
    def test_daa(self):
        daa_base = os.path.join("diamond", "daa", "blast.daa")
        for folder in ["sanitised", "uniprot"]:
            targets = os.path.join(self.master, folder, "uniprot.fasta")
            self.assertTrue(os.path.exists(targets))
            daa = os.path.join(self.master, folder, daa_base)
            self.assertTrue(os.path.exists(daa))
            with tempfile.TemporaryDirectory() as out_folder:
                conf = MikadoConfiguration()
                out_db = os.path.join(out_folder, conf.db_settings.db)
                conf.db_settings.db = out_db
                conf.serialise.files.output_dir = out_folder
                conf.serialise.files.blast_targets = [targets]
                conf.serialise.files.transcripts = self.queries
                conf.serialise.files.xml = [daa]
                logger = create_default_logger(f"test_daa_{folder}")
                self.assertTrue(tempfile.gettempdir() in out_db)
                self.run_loading(daa, out_db, logger, conf)

    def test_xml(self):
        xml_base = os.path.join("xml", "blast.xml.gz")
        for folder in ["sanitised", "uniprot"]:
            for subfolder in ["blast", "blast_parse_seqids", "diamond"]:
                targets = os.path.join(self.master, folder, "uniprot.fasta")
                self.assertTrue(os.path.exists(targets))
                xml = os.path.join(self.master, folder, subfolder, xml_base)
                self.assertTrue(os.path.exists(xml))
                with tempfile.TemporaryDirectory() as out_folder:
                    conf = MikadoConfiguration()
                    conf.serialise.files.output_dir = out_folder
                    out_db = os.path.join(out_folder, conf.db_settings.db)
                    conf.db_settings.db = out_db
                    conf.serialise.files.blast_targets = [targets]
                    conf.serialise.files.transcripts = self.queries
                    conf.serialise.files.xml = [xml]
                    logger = create_default_logger(f"test_xml_{folder}_{subfolder}")
                    self.assertTrue(tempfile.gettempdir() in out_db)
                    self.run_loading(xml, out_db, logger, conf)

    def test_tsv(self):
        tsv_base = os.path.join("tsv", "blast.tsv.gz")
        for folder in ["uniprot", "sanitised"]:
            for subfolder in ["blast", "blast_parse_seqids", "diamond"]:
                targets = os.path.join(self.master, folder, "uniprot.fasta")
                self.assertTrue(os.path.exists(targets))
                tsv = os.path.join(self.master, folder, subfolder, tsv_base)
                self.assertTrue(os.path.exists(tsv))
                with tempfile.TemporaryDirectory() as out_folder:
                    conf = MikadoConfiguration()
                    conf.serialise.files.output_dir = out_folder
                    conf.serialise.files.blast_targets = [targets]
                    conf.serialise.files.transcripts = self.queries
                    conf.serialise.files.xml = [tsv]
                    out_db = os.path.join(out_folder, conf.db_settings.db)
                    conf.db_settings.db = out_db
                    self.assertTrue(tempfile.gettempdir() in out_db)
                    logger = create_default_logger(f"test_tsv_{folder}_{subfolder}", level="DEBUG")
                    self.run_loading(tsv, out_db, logger, conf)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# coding: utf-8

"""
This module defines the functionalities needed to verify the integrity and completeness
of Mikado configuration files. Missing values are replaced with default ones,
while existing values are checked for type and consistency.
"""
import dataclasses
import json
import os.path
import pprint
import marshmallow
from multiprocessing import get_start_method
from logging import Logger
import yaml
from pkg_resources import resource_stream
from ..exceptions import InvalidConfiguration
from ..utilities.log_utils import create_default_logger
import random
import toml
from .configuration import MikadoConfiguration
from .daijin_configuration import DaijinConfiguration
from typing import Union
try:
    from yaml import CSafeLoader as yLoader
except ImportError:
    from yaml import SafeLoader as yLoader
from argparse import Namespace


__author__ = "Luca Venturini"


def create_cluster_config(config: Union[MikadoConfiguration, DaijinConfiguration], args: Namespace, logger: Logger):

    """
    Function to create a configuration for the scheduler for snakemake. It will contain information regarding
    queues, memory, threads etc. per program call.
    :param config: Configuration object
    :param args: namespace with the parameters to use for this run
    :param logger: logger for this run
    :return:
    """

    if (config.scheduler and config.scheduler != "local") or (not config.scheduler and args.cluster_config):
        if not args.queue:
            error = "A queue must be specified for the analysis when in HPC mode. Please relaunch."
            logger.error(error)
            exit(1)
        if args.cluster_config is not None:
            cluster_config = args.cluster_config
        else:
            cluster_config = "daijin_hpc.yaml"
        with open(cluster_config, "wb") as out, resource_stream("Mikado.daijin", "hpc.yaml") as original:
            for pos, line in enumerate(original):
                out.write(line)
                if pos == 0:
                    out.write("    queue: {}\n".format(args.queue).encode())


def check_db(configuration: Union[MikadoConfiguration, DaijinConfiguration]) -> Union[MikadoConfiguration,
                                                                                      DaijinConfiguration]:

    """
    Function to check the validity of the database options.
    :param configuration: (MikadoConfiguration|DaijinConfiguration)
    :return: 
    """

    if configuration.db_settings.dbtype in ("mysql", "postgresql"):
        if configuration.db_settings.dbhost is None:
            raise InvalidConfiguration(
                "No host specified for the {0} database!".format(
                    configuration.db_settings.dbtype))
        if configuration.db_settings.dbuser is None:
            raise InvalidConfiguration(
                "No user specified for the {0} database!".format(
                    configuration.db_settings.dbtype))
        if configuration.db_settings.dbport == 0:
            if configuration.db_settings.dbtype == "mysql":
                configuration.db_settings.dbport = 3306
            else:
                configuration.db_settings.dbport = 5432
    else:
        if configuration.db_settings.db is not None and not os.path.isabs(configuration.db_settings.db):
            if os.path.exists(os.path.join(os.getcwd(), configuration.db_settings.db)):
                configuration.db_settings.db = os.path.join(os.getcwd(), configuration.db_settings.db)
            elif os.path.exists(
                    os.path.join(os.path.dirname(configuration.filename or ""), configuration.db_settings.db)):
                configuration.db_settings.db = os.path.join(
                    os.path.dirname(configuration.filename or ""), configuration.db_settings.db)
            elif os.path.exists(os.path.join(os.path.dirname(configuration.filename or ""),
                                             configuration.db_settings.db)):
                configuration.db_settings.db = os.path.join(os.path.dirname(configuration.filename or ""),
                                                            configuration.db_settings.db)
            else:
                pass

    return configuration


def check_and_load_scoring(configuration: Union[DaijinConfiguration, MikadoConfiguration],
                           logger=None) -> Union[DaijinConfiguration, MikadoConfiguration]:

    """
    Wrapper for the various checks performed on the configuration file.
    :param configuration: the configuration object to finalise
    :type configuration: (MikadoConfiguration|DaijinConfiguration)
    :param logger: external logger instance
    :type logger: (None|Logger)
    :return configuration
    :rtype: (MikadoConfiguration|DaijinConfiguration)
    """

    if not isinstance(logger, Logger):
        logger = create_default_logger("check_json")

    try:
        configuration.load_scoring(logger=logger)
        configuration.check(logger=logger)
        configuration = check_db(configuration)
        if not configuration.multiprocessing_method:
            configuration.multiprocessing_method = get_start_method()
    except InvalidConfiguration as exc:
        logger.exception(exc)
        raise

    assert configuration.seed is not None
    random.seed(configuration.seed % (2 ** 32 - 1))
    return configuration


def load_and_validate_config(raw_configuration: Union[None, MikadoConfiguration, DaijinConfiguration,
                             str, dict], logger=None, external=False) -> Union[MikadoConfiguration, DaijinConfiguration]:
    """
    Function to serialise the JSON for configuration and check its consistency.

    :param raw_configuration: either the file name of the configuration or an initialised object to check and finalise.
    :type raw_configuration: (str | None | dict | MikadoConfiguration | DaijinConfiguration)

    :param external: boolean. If True, presume the file might be an old/incomplete configuration to pass in values.
    Therefore, accept also *partial* configuration files.
    :type external: bool

    :param logger: optional logger to be used.
    :type logger: Logger

    :rtype: (MikadoConfiguration|DaijinConfiguration)
    """

    if not isinstance(logger, Logger):
        logger = create_default_logger("to_json")

    try:
        if isinstance(raw_configuration, (MikadoConfiguration, DaijinConfiguration)):
            config = raw_configuration
        elif isinstance(raw_configuration, dict):
            try:
                config = MikadoConfiguration.Schema().load(raw_configuration, partial=external)
            except marshmallow.exceptions.ValidationError:
                config = DaijinConfiguration.Schema().load(raw_configuration, partial=external)
        elif raw_configuration is None or raw_configuration == '' or raw_configuration == dict():
            config = MikadoConfiguration()
        else:
            assert isinstance(raw_configuration, str), raw_configuration
            raw_configuration = os.path.abspath(raw_configuration)
            if not os.path.exists(raw_configuration) or os.stat(raw_configuration).st_size == 0:
                raise InvalidConfiguration("JSON file {} not found!".format(raw_configuration))
            with open(raw_configuration) as json_file:
                # YAML *might* try to load up the file as the proper object
                if raw_configuration.endswith(".yaml"):
                    config = yaml.load(json_file, Loader=yaml.Loader)
                elif raw_configuration.endswith(".toml"):
                    config = toml.load(json_file)
                elif raw_configuration.endswith(".json"):
                    config = json.loads(json_file.read())
                else:
                    config = toml.load(json_file)
            assert isinstance(config, dict), (config, type(config))
            config["filename"] = raw_configuration
            try:
                config = MikadoConfiguration.Schema().load(config, partial=external)
            except marshmallow.exceptions.ValidationError as mikado_exc:
                try:
                    config = DaijinConfiguration.Schema().load(config, partial=external)
                except marshmallow.exceptions.ValidationError as exc:

                    exc = marshmallow.exceptions.ValidationError(
                        "The configuration file is invalid. \n"
                        f"Validation errors if a Daijin configuration file was expected:\n{exc}\n"
                        f"Validation errors if a Mikado configuration file was expected:\n{mikado_exc}"
                    )
                    logger.critical(exc)
                    raise exc

            config.filename = raw_configuration
            if config.scoring is None:
                check_and_load_scoring(config, logger=logger)
            config.scoring.check(config.pick.orf_loading.minimal_orf_length)

        assert isinstance(config, (MikadoConfiguration, DaijinConfiguration)), type(config)
        config = check_and_load_scoring(config, logger=logger)
    except KeyboardInterrupt:
        raise
    except Exception as exc:
        logger.exception("Loading the configuration file failed with error:\n%s\n\n\n", exc)
        raise InvalidConfiguration(f"The configuration file passed is invalid. Please double check. Exception: {exc}")

    random.seed(config.seed % (2 ** 32 - 1))

    return config



"""
Generic utilities used for BLAST serialising into a DB.
"""

import numpy as np

from .query import id_pattern
from ...parsers.blast_utils import merge
from .aln_string_parser import prepare_aln_strings


def get_off_by_one(record):
    """"""
    ref = getattr(record, "reference", None)
    if ref is None:
        ref = record.__dict__.get("reference", None)
    if ref is None:
        return False
    elif isinstance(ref, str) and "diamond":
        version = record.version
        try:
            macro, minor, micro = [int(_) for _ in version.split(".")]
        except (ValueError, TypeError):
            return False
        if macro > 0 or minor > 9 or (micro > 30 or micro < 25):
            return False
        else:
            return True


def get_multipliers(record, application=None):
    """
    Private quick method to determine the multipliers for a BLAST alignment
    according to the application present in the record.
    :param record:
    :type record: Bio.SearchIO.Record
    :return:
    """

    q_mult, h_mult = 1, 1

    if record is not None:
        application = record.program.upper()
    else:
        application = application.upper()
    if application in ("BLASTN", "TBLASTX", "BLASTP"):
        q_mult = 1
        h_mult = 1
    elif application == "BLASTX":
        q_mult = 3
        h_mult = 1
    elif application == "TBLASTN":
        q_mult = 1
        h_mult = 3

    return q_mult, h_mult


def _get_query_for_blast(record, cache):
    """ This private method formats the name of the query
    recovered from the BLAST hit. It will cause an exception if the target is not
    present in the dictionary.
    :param record: Bio.SearchIO.Record
    :return: current_query (ID in the database), name
    """

    if id_pattern.search(record.id):
        name = id_pattern.search(record.id).groups()[0]
    elif len(record.id.split()) > 1:
        name = record.id.split()[0]
    else:
        name = record.id
    if name in cache:
        return cache[name], name, cache
    else:
        raise ValueError("{} not found (Accession: {})".format(record.id, record.__dict__))


def _get_target_for_blast(alignment, cache):
    """ This private method retrieves the correct target_id
    key for the target of the BLAST. If the entry is not present
    in the database, it will be created on the fly.
    The method returns the index of the current target and
    and an updated target dictionary.
    :param alignment: an alignment child of a BLAST record object
    :type alignment: Bio.SearchIO.Hit
    :return: current_target (ID in the database), targets
    """

    accession = alignment.accession
    al_id = alignment.id

    if id_pattern.search(accession):
        accession = id_pattern.search(accession).groups()[0]
    elif len(accession.split()) > 1:
        accession = accession.split()[0]
    else:
        pass

    if accession in cache:
        return cache[accession], cache
    else:
        if id_pattern.search(al_id):
            al_id = id_pattern.search(al_id).groups()[0]
        elif len(al_id.split()) > 1:
            al_id = al_id.split()[0]
        else:
            pass
        if al_id in cache:
            return cache[al_id], cache
        else:
            raise ValueError("{} not found (Accession: {})".format(alignment.id, alignment.__dict__))


def _np_grouper(data):
    return np.array(np.split(data, np.where(np.diff(data) != 1)[0] + 1))



"""
Basic module with the Query serialiser.
"""

import collections
from sqlalchemy import Column, String, Integer
from ...utilities.dbutils import DBBASE
import re
__author__ = 'Luca Venturini'


id_pattern = re.compile(r"^[^|]*\|([^|]*)\|.*")

# These two classes are OK like this, they do not need more public methods!
# pylint: disable=too-few-public-methods
class Query(DBBASE):
    """
    Very simple serialization class for Query objects.

    :return id: integer key
    :rtype id: int

    :return name: name of the queries
    :rtype name: str

    :return length: length of the queries
    :rtype length: int
    """

    __tablename__ = "query"
    query_id = Column(Integer, primary_key=True)
    query_name = Column(String(200), unique=True, index=True)
    # This so we can load data also from the orf class
    query_length = Column(Integer, nullable=True)

    named_tup = collections.namedtuple("Query",
                                       ["query_id", "query_name", "query_length"])

    def __init__(self, name, length):
        if not isinstance(name, str):
            raise TypeError("Invalid name: {0}".format(name))
        if not isinstance(length, int) or length <= 0:
            raise TypeError("Invalid length value: {0}".format(length))

        if id_pattern.search(name):
            name = id_pattern.search(name).groups()[0]

        self.query_name = name
        self.query_length = length

    def as_tuple(self):
        """Quick function to convert the SQLalchemy object into
        a named tuple with the same fields"""
        return self.named_tup(self.query_id, self.query_name, self.query_length)

    def as_dict(self):
        return {
            "query_name": self.query_name,
            "query_length": self.query_length
        }



import shutil

from ...parsers.GFF import GFF3
from .indexing import create_index, check_index
from ...exceptions import CorruptIndex
import os
import tempfile

from ...utilities import create_shm_handle


def prepare_index(args, queue_logger):
    index_name = os.path.abspath("{0}.midx".format(args.reference.name))
    ref_gff = isinstance(args.reference, GFF3)

    if args.shm is True:
        shm_handle = create_shm_handle()
    else:
        shm_handle = tempfile.NamedTemporaryFile(suffix=".midx")

    if args.index is True:
        create_index(args.reference, queue_logger=queue_logger, index_name=index_name,
                     ref_gff=ref_gff, use_shm=args.shm,
                     exclude_utr=False, protein_coding=False)
        assert os.path.exists(index_name), \
            "Index {} should have been created but is now absent! File system problems?".format(index_name)
        return index_name

    if os.path.exists(index_name) and os.stat(args.reference.name).st_mtime >= os.stat(index_name).st_mtime:
        queue_logger.warning("Reference index obsolete, deleting and rebuilding.")
        try:
            os.remove("{0}.midx".format(args.reference.name))
        except (OSError, PermissionError) as exc:
            queue_logger.error(
                "I cannot delete the old index due to permission errors. I will create a temporary one instead.")
            __index = tempfile.NamedTemporaryFile(suffix=".midx")
            index_name = __index.name

        ref_gff = isinstance(args.reference, GFF3)
        create_index(args.reference, queue_logger, index_name, ref_gff=ref_gff,  use_shm=args.shm,
                     protein_coding=args.protein_coding, exclude_utr=args.exclude_utr)
    elif os.path.exists(index_name):
        # queue_logger.info("Starting loading the indexed reference")
        queue_logger.info("Index found")
        try:
            check_index(args.reference.name, queue_logger)
            queue_logger.info("Index valid, proceeding.")
        except CorruptIndex as exc:
            queue_logger.warning(exc)
            queue_logger.warning("Reference index corrupt, deleting and rebuilding.")
            try:
                os.remove("{0}.midx".format(args.reference.name))
            except (OSError, PermissionError) as exc:
                queue_logger.error(
                    "I cannot delete the old index due to permission errors. I will create a temporary one instead.")
                __index = tempfile.NamedTemporaryFile(suffix=".midx")
                index_name = __index.name
            create_index(args.reference, queue_logger, index_name, ref_gff=ref_gff, use_shm=args.shm,
                         protein_coding=args.protein_coding, exclude_utr=args.exclude_utr)
    else:
        if args.no_save_index is True:
            __index = tempfile.NamedTemporaryFile(suffix=".midx", delete=False)
            index_name = __index.name
        create_index(args.reference, queue_logger, index_name, ref_gff=ref_gff, use_shm=args.shm,
                     protein_coding=args.protein_coding, exclude_utr=args.exclude_utr)
        assert os.path.exists(index_name), "Index file {} should have been created but is now absent!".format(
            index_name)

    try:
        if shm_handle is not None:
            shm_handle.close()
            queue_logger.info(f"Copying the database into the temporary index {shm_handle.name}")
            shutil.copy(index_name, shm_handle.name)
            index_name = shm_handle.name
            queue_logger.info(f"Copied the database into the temporary index {shm_handle.name}")
    except (PermissionError, FileNotFoundError, OSError, TypeError, ValueError) as exc:
        queue_logger.error("Failed to copy the database to SHM. Using the in-disk copy.")
        queue_logger.error(f"Specific exception: {exc}")

    assert os.path.exists(index_name), (args.reference, index_name)
    return index_name



#!/usr/bin/env python3

import sys, argparse, os
# from Bio import SeqIO
from io import IOBase
import pyfaidx
import textwrap

def main():

    def check_type(string):
        if os.path.exists(string) and not os.path.isdir(string): return open(string)
        else: return set(string.split(','))

    parser=argparse.ArgumentParser(description='A simple script that retrieves the FASTA sequences from a file given a list of ids.')
    parser.add_argument("-v", "--reverse", action="store_true", default=False, help="Retrieve entries which are not in the list, as in grep -v (a homage).")
    parser.add_argument('list', type=check_type, help='File with the list of the ids to recover, one by line. Alternatively, names separated by commas.')
    parser.add_argument('fasta', type=argparse.FileType('r'), help='FASTA file.')
    parser.add_argument('out', type=argparse.FileType('w'), help='Optional output file.', nargs='?', default=sys.stdout)
    args=parser.parse_args()

    if isinstance(args.list, IOBase):
        ids = set([line.rstrip() for line in args.list.readlines()])
    else:
        ids=args.list

    args.fasta.close()
    fasta = pyfaidx.Fasta(args.fasta.name)

    for name in ids:
        assert name in fasta
        print(">{0}".format(name), file=args.out)
        print(*textwrap.wrap(str(fasta[name]), width=60),
              sep="\n", file=args.out)
        
    # for record in SeqIO.parse(args.fasta, 'fasta'):
    #     if record.id in ids:
    #         if args.reverse: continue
    #         print(record.format('fasta'), file=args.out, end='')
    #         ids.remove(record.id)
    #         if len(ids)==0 and not args.reverse: break
    #     else:
    #         if args.reverse:
    #             print(record.format('fasta'), file=args.out, end='')

if __name__=='__main__': main()



from ...exceptions import InvalidTranscript
from ...transcripts import Transcript
import functools


def parse_prediction_bed12(args, queue_logger):
    """"""

    constructor = functools.partial(Transcript, logger=queue_logger, trust_orf=True, accept_undefined_multi=True)
    transcript = None
    invalids = set()
    for row in args.prediction:
        if row.header is True:
            continue
        yield transcript
        try:
            transcript = constructor(row)
        except (InvalidTranscript, AssertionError, TypeError, ValueError):
            queue_logger.warning("Row %s is invalid, skipping.", row)
            transcript = None
            invalids.add(row.id)
            continue
        transcript.parent = transcript.id

    yield transcript



import csv
from ..transcripts.transcriptcomputer import TranscriptComputer
from ..loci import Gene
from ..parsers.GFF import GFF3
from ..utilities.log_utils import create_default_logger
from collections import defaultdict
import numpy as np
from collections import Counter
# pylint: disable=E1101
np.seterr(all="ignore")  # Suppress warnings
np.warnings.filterwarnings("ignore")
# pylint: enable=E1101


def itemize(counter):
    """
    Private static method to convert a counter into a 2-dimensional numpy array.
    :param counter: the counter to transform
    :type counter: dict
    :return: Numpy array of the counter
    :rtype: array
    """

    if len(counter) == 0:
        return np.array([[], []])  # Empty 2dimensional array
    else:
        return np.array(list(zip(*counter.items())))


def weighted_mode(a, w, *, axis=0):
    """Returns an array of the weighted modal (most common) value in a.
    If there is more than one such value, only the first is returned.
    The bin-count for the modal bins is also returned.
    This is an extension of the algorithm in scipy.stats.mode.
    Copied from sklearn source code on 26/01/2021.
    Parameters
    ----------
    a : array-like
        n-dimensional array of which to find mode(s).
    w : array-like
        n-dimensional array of weights for each value.
    axis : int, default=0
        Axis along which to operate. Default is 0, i.e. the first axis.
    Returns
    -------
    vals : ndarray
        Array of modal values.
    score : ndarray
        Array of weighted counts for each mode.
    Examples
    --------
    >>> x = [4, 1, 4, 2, 4, 2]
    >>> weights = [1, 1, 1, 1, 1, 1]
    >>> weighted_mode(x, weights)
    (array([4.]), array([3.]))
    The value 4 appears three times: with uniform weights, the result is
    simply the mode of the distribution.
    >>> weights = [1, 3, 0.5, 1.5, 1, 2]  # deweight the 4's
    >>> weighted_mode(x, weights)
    (array([2.]), array([3.5]))
    The value 2 has the highest score: it appears twice with weights of
    1.5 and 2: the sum of these is 3.5.
    See Also
    --------
    scipy.stats.mode
    """
    if axis is None:
        a = np.ravel(a)
        w = np.ravel(w)
        axis = 0
    else:
        a = np.asarray(a)
        w = np.asarray(w)

    if a.shape != w.shape:
        w = np.full(a.shape, w, dtype=w.dtype)

    scores = np.unique(np.ravel(a))       # get ALL unique values
    testshape = list(a.shape)
    testshape[axis] = 1
    oldmostfreq = np.zeros(testshape)
    oldcounts = np.zeros(testshape)
    for score in scores:
        template = np.zeros(a.shape)
        ind = (a == score)
        template[ind] = w[ind]
        counts = np.expand_dims(np.sum(template, axis), axis)
        mostfrequent = np.where(counts > oldcounts, score, oldmostfreq)
        oldcounts = np.maximum(counts, oldcounts)
        oldmostfreq = mostfrequent
    return mostfrequent, oldcounts


def weighted_percentile(a, percentile=np.array([75, 25]), weights=None):
    """
    O(nlgn) implementation for weighted_percentile.
    From: http://stackoverflow.com/questions/21844024/weighted-percentile-using-numpy
    Kudos to SO user Nayyary http://stackoverflow.com/users/2004093/nayyarv

    :param a: array
    :type a: (dict|list|np.array|set|tuple)

    :param percentile: the percentiles to calculate.
    :type percentile: (np.array|list|tuple)

    """

    percentile = np.array(percentile)/100.0

    if isinstance(a, (dict, Counter)):
        a, weigths = itemize(a)
    else:
        assert isinstance(a, (list, set, tuple, np.ndarray)), (a, type(a))
        if not isinstance(a, type(np.array)):
            a = np.array(a)
        if weights is None:
            weights = np.ones(len(a))

    a_indsort = np.argsort(a)
    a_sort = a[a_indsort]
    weights_sort = weights[a_indsort]
    ecdf = np.cumsum(weights_sort)

    percentile_index_positions = percentile * (weights.sum() - 1) + 1
    # need the 1 offset at the end due to ecdf not starting at 0
    locations = np.searchsorted(ecdf, percentile_index_positions)

    out_percentiles = np.zeros(len(percentile_index_positions))

    for i, empiricalLocation in enumerate(locations):
        # iterate across the requested percentiles
        if ecdf[empiricalLocation-1] == np.floor(percentile_index_positions[i]):
            # i.e. is the percentile in between 2 separate values
            uppWeight = percentile_index_positions[i] - ecdf[empiricalLocation-1]
            lowWeight = 1 - uppWeight

            out_percentiles[i] = a_sort[empiricalLocation-1] * lowWeight + \
                                 a_sort[empiricalLocation] * uppWeight
        else:
            # i.e. the percentile is entirely in one bin
            out_percentiles[i] = a_sort[empiricalLocation]

    return out_percentiles


class Calculator:

    """
    This class has the purpose of parsing a reference file,
    calculating the statistics, and printing them out.
    """

    def __init__(self, parsed_args):

        """Constructor function"""

        self.gff = parsed_args.gff

        self.atype = self.gff.__annot_type__

        self.__logger = create_default_logger("calculator")
        if parsed_args.verbose is True:
            self.__logger.setLevel("DEBUG")
        self.only_coding = parsed_args.only_coding
        self.tab_handle = parsed_args.tab_stats
        if self.tab_handle is not None:
            __fieldnames = ["TID", "GID",
                            "Coordinates",
                            "Strand",
                            "Exon number",
                            "cDNA length",
                            "Intronic length",
                            "CDS length",
                            "# coding exons",
                            "cDNA/CDS ratio",
                            "5'UTR",
                            "5'UTR exons",
                            "3'UTR",
                            "3'UTR exons"]
            self.tab_writer = csv.DictWriter(self.tab_handle, __fieldnames, delimiter="\t")
            self.tab_writer.writeheader()

        self.out = parsed_args.out
        self.genes = dict()
        self.coding_genes = []
        self.__distances = np.array([])
        self.__positions = defaultdict(list)
        self.__coding_positions = defaultdict(list)
        self.__coding_distances = np.array([])
        self.__fieldnames = ['Stat', 'Total', 'Average', 'Mode', 'Min',
                             '1%', '5%', '10%', '25%', 'Median', '75%', '90%', '95%', '99%', 'Max']
        self.__rower = csv.DictWriter(self.out,
                                      self.__fieldnames,
                                      delimiter="\t")
        self.__arrays = dict()
        self.__stores = dict()
        self.__prepare_stores()

    def parse_input(self):
        """
        Method to parse the input GTF/GFF file.
        """
        transcript2gene = dict()

        derived_features = set()

        genes = dict()
        orphaned = set()
        orphan_counter = 0

        for record in self.gff:
            if self.atype == "bed12":
                if not record.parent:
                    record.parent = "{}.gene".format(record.id)
                current_gene = Gene(record, gid=record.parent[0], only_coding=self.only_coding,
                                    logger=self.__logger, use_computer=True)
                transcript2gene[record.id] = record.parent[0]
                current_gene.transcripts[record.id] = TranscriptComputer(record, logger=self.__logger)
                assert current_gene.id not in genes, f"Duplicated ID found: {current_gene.id}"
                genes[current_gene.id] = current_gene
            elif self.atype == "bam":
                transcript = TranscriptComputer(record, logger=self.__logger)
                transcript.parent = transcript.id + "_gene"
                gcounter = 0
                while transcript.parent[0] in genes:
                    gcounter += 1
                    transcript.parent = transcript.id + f"_gene.{gcounter}"

                genes[transcript.parent[0]] = Gene(transcript, gid=transcript.parent[0],
                                                   only_coding=self.only_coding, logger=self.__logger,
                                                   use_computer=True)
            elif record.is_gene is True:
                if record.id in genes:  # Necessary for unsorted inputs
                    genes[record.id].attributes.update(record.attributes)
                    genes[record.id].start, genes[record.id].end = record.start, record.end
                    assert (genes[record.id].strand, genes[record.id].chrom) == record.strand, record.chrom
                else:
                    gene = Gene(record, only_coding=self.only_coding, logger=self.__logger, use_computer=True)
                    genes[gene.id] = gene
            elif record.is_transcript is True or (self.atype == GFF3.__annot_type__ and record.feature == "match"):
                if record.parent is None and record.feature != "match":
                    raise TypeError("No parent found for:\n{0}".format(str(record)))
                elif record.feature == "match":
                    gid = record.id
                elif len(record.parent) == 0:
                    self.__logger.warning("No gene found for %s, creating a mock one.", record.id)
                    gid = record.id + ".gene"
                else:
                    gid = record.parent[0]
                if gid not in genes:
                    gene = Gene(
                        record,
                        gid=gid,
                        only_coding=self.only_coding,
                        logger=self.__logger,
                        use_computer=True)
                    genes[gid] = gene

                transcript2gene[record.id] = gid
                if record.id not in genes[gid]:
                    genes[gid].transcripts[record.id] = TranscriptComputer(record, logger=self.__logger)
                else:
                    genes[gid].transcripts[record.id].start, genes[gid].transcripts[record.id].end = (record.start,
                                                                                                      record.end)
                    genes[gid].transcripts[record.id].attributes.update(record.attributes)

            elif record.is_derived is True and record.is_exon is False:
                derived_features.add(record.id)
            elif record.is_exon is True:
                if self.atype != GFF3.__annot_type__:
                    if record.gene not in genes:
                        gene = Gene(
                            record,
                            only_coding=self.only_coding,
                            logger=self.__logger,
                            use_computer=True)
                        genes[record.gene] = gene
                        transcript2gene[record.transcript] = record.gene

                    transcript2gene[record.transcript] = record.gene
                    if record.transcript not in genes[record.gene].transcripts:
                        genes[record.gene].transcripts[record.transcript] = TranscriptComputer(record,
                                                                                         logger=self.__logger)
                    else:
                        genes[record.gene].transcripts[record.transcript].add_exon(record)

                elif self.atype == GFF3.__annot_type__ and "cDNA_match" in record.feature:
                    # Here we assume that we only have "cDNA_match" lines, with no parents
                    record.parent = record.id
                    if record.id not in genes:
                        genes[record.id] = Gene(
                            record,
                            gid=record.id,
                            only_coding=self.only_coding,
                            logger=self.__logger,
                            use_computer=True)
                    transcript2gene[record.id] = record.id
                    if record.id not in genes[record.id]:
                        genes[record.id].transcripts[record.id] = TranscriptComputer(record,
                                                                                     logger=self.__logger)
                    else:
                        genes[record.id].add_exon(record)
                elif self.atype == GFF3.__annot_type__:
                    found = False
                    for parent in record.parent:
                        if parent in transcript2gene:
                            gene_id = transcript2gene[parent]
                            genes[gene_id].add_exon(record)
                            found = True
                            break
                    if found is False:
                        orphan_counter += 1
                        self.__logger.debug(f"Storing orphan line {record.parent}, {record.start}, {record.end}")
                        orphaned.add(record)
            elif record.header is True:
                continue
            else:
                continue

        assert orphan_counter == len(orphaned), f"We have lost lines! Expected {orphan_counter}, found {len(orphaned)}"
        self.__logger.debug(f"{len(orphaned)} orphan lines.")
        for orphan in orphaned:
            found = False
            for parent in orphan.parent:
                if found is True:
                    continue
                if parent in transcript2gene:
                    gene_id = transcript2gene[parent]
                    genes[gene_id].add_exon(orphan)
                    found = True
                    self.__logger.debug(f"Found the gene for the record of {parent} (start {orphan.start}, "
                                        f"end {orphan.end})")
            if found is False:  # Create a mock gene
                self.__logger.debug(f"Line for record {orphan.parent} ({orphan.start}, {orphan.end}) is still "
                                    f"orphaned")
                transcript = TranscriptComputer(orphan, logger=self.__logger)
                transcript.parent = transcript.id + "_gene"
                gene = Gene(transcript, only_coding=self.only_coding,
                            logger=self.__logger,
                            use_computer=True)
                genes[gene.id] = gene

        for gene_id, gene in genes.items():
            self.__store_gene(gene)

    def __call__(self):

        self.parse_input()
        self.gff.close()
        distances = []
        for chromosome in self.__positions:
            __ordered = sorted(self.__positions[chromosome])
            for index, position in enumerate(__ordered[:-1]):
                distances.append(__ordered[index + 1][0] - position[1])

        self.__distances = np.array(distances)

        distances = []
        for chromosome in self.__positions:
            __ordered = sorted(self.__coding_positions[chromosome])
            for index, position in enumerate(__ordered[:-1]):
                distances.append(__ordered[index + 1][0] - position[1])

        self.__coding_distances = np.array(distances)
        self.writer()
        self.out.close()

    @staticmethod
    def get_stats(row: dict, array: np.array) -> dict:
        """
        Method to calculate the necessary statistic from a row of values.
        :param row: the output dictionary row.
        :type row: dict

        :param array: an array of values.

        :rtype : dict
        """

        # Decimal to second digit precision

        if array is None or len(array) == 0:
            return row

        array, weights = array

        if len(weights) > 0 and sum(weights) != 0:
            row["Average"] = "{0:,.2f}".format(round(
                sum(_[0] * _[1] for _ in zip(array, weights)) / sum(weights), 2))

            mode = weighted_mode(array, weights)
            # mode = scipy.stats.mode(array, axis=None)
            row["Mode"] = mode[0][0]
        else:
            row["Average"] = "NA"
            row["Mode"] = "NA"

        keys = ['Min', '1%', '5%', '10%', '25%', 'Median', '75%', '90%', '95%', '99%', 'Max']
        if len(array) == 0:
            quantiles = ["NA"]*len(keys)
        else:
            quantiles = weighted_percentile(array,
                                            [0, 1, 5, 10, 25, 50, 75, 90, 95, 99, 100],
                                            weights)
        for key, val in zip(keys, quantiles):
            try:
                row[key] = "{0:,.0f}".format(val)  # No decimal
            except KeyError:
                row[key] = val
            except ValueError:
                row[key] = val
            except TypeError:
                row[key] = val
            except Exception:
                raise
        return row

    def __prepare_stores(self):

        self.__stores["genes"] = set()
        self.__stores["coding_genes"] = set()
        self.__stores["monoexonic_genes"] = set()
        self.__stores["transcripts_per_gene"] = dict()
        self.__stores["coding_transcripts_per_gene"] = dict()
        self.__stores["exons"], self.__stores["exons_coding"] = dict(), dict()
        self.__stores["exon_num"], self.__stores["exon_num_coding"] = dict(), dict()
        self.__stores["introns"], self.__stores["introns_coding"] = dict(), dict()
        self.__stores["cds_introns"] = dict()
        self.__stores["cds_exons"] = dict()
        self.__stores["cds_exon_num"] = dict()
        self.__stores["cds_exon_num_coding"] = dict()
        self.__stores["cdna_lengths"] = dict()  # Done
        self.__stores["cdna_lengths_coding"] = dict()
        self.__stores["cds_lengths"] = dict()  # Done
        self.__stores["cds_lengths_coding"] = dict()  # Done
        self.__stores["cds_ratio"] = dict()
        self.__stores["monoexonic_lengths"] = dict()
        self.__stores["multiexonic_lengths"] = dict()
        self.__stores["monocds_lengths"] = dict()

        self.__stores["five_utr_lengths"] = dict()
        self.__stores["five_utr_nums"] = dict()
        self.__stores["three_utr_lengths"] = dict()
        self.__stores["three_utr_nums"] = dict()
        self.__stores["end_distance_from_junction"] = dict()

    def __store_gene(self, gene):

        """

        :param gene:
        :type gene: (None|Mikado.loci.Gene)
        :return:
        """

        if gene is None:
            return

        gene.finalize()
        if self.only_coding is True and gene.is_coding is False:
            return
        if len(gene.transcripts) == 0:
            self.__logger.warning("Gene %s only has invalid, or no, transcripts. Ignoring it.", gene.id)
            return

        self.__positions[gene.chrom].append((gene.start, gene.end, str(gene.strand)))
        self.__stores["genes"].add(gene.id)
        if gene.is_coding is True:
            self.__coding_positions[gene.chrom].append((gene.start, gene.end, gene.strand))
            self.__stores["coding_genes"].add(gene.id)

        self.__stores["transcripts_per_gene"][gene.num_transcripts] = self.__stores["transcripts_per_gene"].get(
            gene.num_transcripts, 0) + 1
        self.__stores["coding_transcripts_per_gene"][
            gene.num_coding_transcripts] = self.__stores["coding_transcripts_per_gene"].get(
            gene.num_coding_transcripts, 0) + 1

        if gene.monoexonic is True:
            self.__stores["monoexonic_genes"].add(gene.id)

        for tid in gene.transcripts:

            if self.tab_handle is not None:
                # __fieldnames = ["TID", "GID",
                #                 "Exon number",
                #                 "cDNA length",
                #                 "CDS length",
                #                 "# coding exons",
                #                 "cDNA/CDS ratio",
                #                 "5'UTR",
                #                 "5'UTR exons",
                #                 "3'UTR",
                #                 "3'UTR exons"]
                row = dict()
                row["TID"] = tid
                row["GID"] = gene.id
                row["Coordinates"] = "{}:{}-{}".format(gene.chrom,
                                                       gene.transcripts[tid].start,
                                                       gene.transcripts[tid].end)
                row["Strand"] = gene.strand if gene.strand else "."
                row["Exon number"] = len(gene.transcripts[tid].exon_lengths)
                row["cDNA length"] = gene.transcripts[tid].cdna_length
                row["Intronic length"] = sum([_[1] - _[0] for _ in gene.transcripts[tid].introns])
                row["CDS length"] = gene.transcripts[tid].selected_cds_length
                row["# coding exons"] = len(gene.transcripts[tid].cds_exon_lengths)
                row["cDNA/CDS ratio"] = "{:.2f}".format(100 * gene.transcripts[tid].selected_cds_length / gene.transcripts[tid].cdna_length)
                row["5'UTR"] = gene.transcripts[tid].five_utr_length
                row["5'UTR exons"] = gene.transcripts[tid].five_utr_num
                row["3'UTR"] = gene.transcripts[tid].three_utr_length
                row["3'UTR exons"] = gene.transcripts[tid].three_utr_num
                self.tab_writer.writerow(row)

            for el in gene.transcripts[tid].exon_lengths:
                self.__stores["exons"][el] = self.__stores["exons"].get(el, 0) + 1
            exon_number = len(gene.transcripts[tid].exon_lengths)
            self.__stores["exon_num"][exon_number] = self.__stores["exon_num"].get(exon_number, 0) + 1
            if exon_number == 1:
                self.__stores["monoexonic_lengths"][
                    gene.transcripts[tid].cdna_length] = self.__stores["monoexonic_lengths"].get(
                    gene.transcripts[tid].cdna_length, 0) + 1
            else:
                self.__stores["multiexonic_lengths"][
                    gene.transcripts[tid].cdna_length] = self.__stores["multiexonic_lengths"].get(
                        gene.transcripts[tid].cdna_length, 0) + 1

            for il in gene.transcripts[tid].intron_lengths:
                self.__stores["introns"][il] = self.__stores["introns"].get(il, 0) + 1

            for cdsil in gene.transcripts[tid].cds_intron_lengths:
                self.__stores["cds_introns"][cdsil] = self.__stores["cds_introns"].get(cdsil, 0) + 1

            for cdsel in gene.transcripts[tid].cds_exon_lengths:
                self.__stores["cds_exons"][cdsel] = self.__stores["cds_exons"].get(cdsel, 0) + 1

            cds_num = len(gene.transcripts[tid].cds_exon_lengths)
            sel_length = gene.transcripts[tid].selected_cds_length
            if (cds_num == 1) or (exon_number == 1 and gene.transcripts[tid].selected_cds_length > 0):
                self.__stores["monocds_lengths"][sel_length] = self.__stores["monocds_lengths"].get(sel_length, 0) + 1

            self.__stores["cds_exon_num"][cds_num] = self.__stores["cds_exon_num"].get(cds_num, 0) + 1
            self.__stores["cdna_lengths"][
                gene.transcripts[tid].cdna_length] = self.__stores["cdna_lengths"].get(
                gene.transcripts[tid].cdna_length, 0) + 1
            self.__stores["cds_lengths"][
                gene.transcripts[tid].selected_cds_length] = self.__stores["cds_lengths"].get(
                gene.transcripts[tid].selected_cds_length, 0) + 1

            if gene.transcripts[tid].selected_cds_length > 0:
                __cds_length = gene.transcripts[tid].selected_cds_length
                __cdna_length = gene.transcripts[tid].cdna_length
                assert __cds_length > 0
                self.__stores["cds_lengths_coding"][
                    gene.transcripts[tid].selected_cds_length] = self.__stores["cds_lengths_coding"].get(
                    gene.transcripts[tid].selected_cds_length, 0) + 1

                for key, attribute in (("five_utr_lengths", "five_utr_length"),
                                       ("three_utr_lengths", "three_utr_length"),
                                       ("five_utr_nums", "five_utr_num"),
                                       ("three_utr_nums", "three_utr_num"),
                                       ("end_distance_from_junction", "end_distance_from_junction")):
                    val = getattr(gene.transcripts[tid], attribute)
                    self.__stores[key][val] = self.__stores[key].get(val, 0) + 1

                cds_ratio = 100 * __cds_length / __cdna_length
                self.__stores["cds_ratio"][cds_ratio] = self.__stores["cds_ratio"].get(cds_ratio, 0) + 1

            if self.only_coding is False:
                if gene.transcripts[tid].selected_cds_length > 0:
                    cdl = gene.transcripts[tid].cdna_length
                    self.__stores["cdna_lengths_coding"][cdl] = self.__stores["cdna_lengths_coding"].get(cdl, 0) + 1
                    for el in gene.transcripts[tid].exon_lengths:
                        self.__stores["exons_coding"][el] = self.__stores["exons_coding"].get(el, 0) + 1
                    num_exons = len(gene.transcripts[tid].exon_lengths)

                    self.__stores["exon_num_coding"][num_exons] = self.__stores["exon_num_coding"].get(num_exons,
                                                                                                       0) + 1
                    cds_num_exons = len(gene.transcripts[tid].cds_exon_lengths)
                    self.__stores["cds_exon_num_coding"][cds_num_exons] = self.__stores[
                                                                              "cds_exon_num_coding"].get(cds_num_exons,
                                                                                                         0) + 1
                    for il in gene.transcripts[tid].intron_lengths:
                        self.__stores["introns_coding"][il] = self.__stores["introns_coding"].get(il, 0) + 1
        return

    def __finalize_arrays(self):

        self.__arrays["Transcripts per gene"] = itemize(self.__stores["transcripts_per_gene"])

        self.__arrays["Coding transcripts per gene"] = itemize(self.__stores["coding_transcripts_per_gene"])
        self.__arrays["Intergenic distances"] = itemize(Counter(self.__distances))
        self.__arrays["Intergenic distances (coding)"] = itemize(Counter(self.__coding_distances))

        self.__arrays['CDNA lengths'] = itemize(self.__stores["cdna_lengths"])
        self.__arrays["CDNA lengths (mRNAs)"] = itemize(self.__stores["cdna_lengths_coding"])
        self.__arrays['CDS lengths'] = itemize(self.__stores["cds_lengths"])
        if self.only_coding is False:
            self.__arrays["CDS lengths (mRNAs)"] = itemize(self.__stores["cds_lengths_coding"])
            self.__arrays['Exons per transcript (mRNAs)'] = itemize(self.__stores["exon_num_coding"])
            self.__arrays['Exon lengths (mRNAs)'] = itemize(self.__stores["exons_coding"])
            self.__arrays["CDS exons per transcript (mRNAs)"] = itemize(
                self.__stores["cds_exon_num_coding"])

        self.__arrays['Monoexonic transcripts'] = itemize(self.__stores["monoexonic_lengths"])
        self.__arrays['MonoCDS transcripts'] = itemize(self.__stores["monocds_lengths"])
        self.__arrays['Exons per transcript'] = itemize(self.__stores["exon_num"])
        self.__arrays['Exon lengths'] = itemize(self.__stores["exons"])
        self.__arrays["Intron lengths"] = itemize(self.__stores["introns"])
        self.__arrays["Intron lengths (mRNAs)"] = itemize(self.__stores["introns_coding"])
        self.__arrays["CDS exons per transcript"] = itemize(self.__stores["cds_exon_num"])
        self.__arrays["CDS exon lengths"] = itemize(self.__stores["cds_exons"])
        self.__arrays["CDS Intron lengths"] = itemize(self.__stores["cds_introns"])
        self.__arrays["5'UTR exon number"] = itemize(self.__stores["five_utr_nums"])
        self.__arrays["3'UTR exon number"] = itemize(self.__stores["three_utr_nums"])
        self.__arrays["5'UTR length"] = itemize(self.__stores["five_utr_lengths"])
        self.__arrays["3'UTR length"] = itemize(self.__stores["three_utr_lengths"])
        self.__arrays["Stop distance from junction"] = itemize(
            self.__stores["end_distance_from_junction"])
        self.__arrays["CDS/cDNA ratio"] = itemize(self.__stores["cds_ratio"])
    # pylint: enable=too-many-locals,too-many-statements

    def writer(self):
        """Method which creates the final output"""

        self.__finalize_arrays()
        self.__rower.writeheader()
        self.__write_statrow('Number of genes',
                             len(self.__stores["genes"]))
        self.__write_statrow("Number of genes (coding)",
                             len(self.__stores["coding_genes"]))
        self.__write_statrow("Number of monoexonic genes",
                             len(self.__stores["monoexonic_genes"])
                             )
        self.__write_statrow('Transcripts per gene',
                             total=np.dot)
        self.__write_statrow("Coding transcripts per gene", total=np.dot)

        self.__write_statrow('CDNA lengths', total=np.dot)
        self.__write_statrow("CDNA lengths (mRNAs)", total=np.dot)
        self.__write_statrow('CDS lengths', total=np.dot)
        if self.only_coding is False:
            self.__write_statrow("CDS lengths (mRNAs)", total=False)

        self.__write_statrow("CDS/cDNA ratio", total=False)

        self.__write_statrow('Monoexonic transcripts', total=sum)
        self.__write_statrow('MonoCDS transcripts', total=sum)
        self.__write_statrow('Exons per transcript', total=np.dot)

        if self.only_coding is False:
            self.__write_statrow('Exons per transcript (mRNAs)',
                                 total="Exon lengths (mRNAs)")
        self.__write_statrow('Exon lengths', total=False)
        if self.only_coding is False:
            self.__write_statrow('Exon lengths (mRNAs)', total=False)
        self.__write_statrow("Intron lengths", total=False)
        if self.only_coding is False:
            self.__write_statrow("Intron lengths (mRNAs)", total=False)

        self.__write_statrow("CDS exons per transcript",
                             total="CDS exon lengths")

        if self.only_coding is False:
            self.__write_statrow("CDS exons per transcript (mRNAs)",
                                 total="CDS exon lengths")

        self.__write_statrow("CDS exon lengths", total=np.dot)
        self.__write_statrow("CDS Intron lengths", total=np.dot)
        self.__write_statrow("5'UTR exon number", total=sum)
        self.__write_statrow("3'UTR exon number", total=sum)
        self.__write_statrow("5'UTR length", total=np.dot)
        self.__write_statrow("3'UTR length", total=np.dot)
        self.__write_statrow("Stop distance from junction",
                             total=False)
        self.__write_statrow("Intergenic distances", total=False)
        self.__write_statrow("Intergenic distances (coding)", total=False)

    def __write_statrow(self, stat, total=True):
        """
        Static method to write out a statistic to the
        output file.
        :param stat: the name of the row
        :type stat: str
        :param total: value to display in the "Total" column
        :type total: str | int | sum | bool
        """
        row = dict()
        for key in self.__fieldnames:
            row[key] = "NA"
        row["Stat"] = stat
        if total is False:
            total = "NA"
        elif total is True:
            if len(self.__arrays[stat]) == 2 and isinstance(
                        self.__arrays[stat], np.ndarray):
                total = len(self.__arrays[stat][0])
            else:
                total = len(self.__arrays[stat])
        else:
            if total is sum:
                try:
                    _, weights = self.__arrays[stat]
                    total = sum(weights)
                except ValueError:
                    total = "NA"
            elif total is np.dot:
                total = np.dot(self.__arrays[stat][0],
                                  self.__arrays[stat][1])

            elif not isinstance(total, int):
                assert total in self.__arrays
                if len(self.__arrays[total]) == 2 and isinstance(
                        self.__arrays[total], np.ndarray):

                    total = len(self.__arrays[total][0])
                else:
                    total = len(self.__arrays[total])
            else:
                pass  # Just keep the total that was passed from the external

        row["Total"] = total
        if stat in self.__arrays:
            current_array = self.__arrays[stat]
        else:
            current_array = None
        row = self.get_stats(row, current_array)
        self.__rower.writerow(row)
# pylint: enable=too-many-instance-attributes



#!/usr/local/bin/env python3

"""
This module describes an abstract class which underlies the parsing of
GTF/GFF files.
"""

import abc
import copy
from sys import intern
import re
from .parser import _attribute_definition


__author__ = 'Luca Venturini'

from Mikado.exceptions import InvalidParsingFormat

[intern(_) for _ in ["+", "-", "?", "true", "True", "false", "False"]]


# This class has exactly how many attributes I need it to have
# pylint: disable=too-many-instance-attributes
class GFAnnotation(metaclass=abc.ABCMeta):

    __negative_order = []
    __positive_order = []

    """
    This abstract class describes a generic GTF/GFF annotation line.
    The parsers for those two type of files inherit from this abstract class,
    which defines common methods and properties.
    """

    @abc.abstractmethod
    def __init__(self, line, my_line='', header=False):
        self.attributes = dict()
        self.__strand = None
        self.header = header
        self.chrom, self.source, self.feature = None, None, None
        # pylint: disable=invalid-name
        self.id = None
        # pylint: enable=invalid-name
        self.parent = []
        self.start, self.end = None, None
        self.__score = None
        self.__phase = None
        self.__frame = None
        self._line = "NA"
        self.__gene = None
        self._transcript = None
        self.__feature = None
        self.attribute_order = []
        if line is None or (line == '' and my_line == ''):  # Empty constructor
            self._fields = []
            self._line = ""
            return
        elif line == '' and my_line != "":
            self._line = my_line
        else:
            self._line = line

        self._line = self._line.strip()

        self._fields = self._line.split('\t')

        if self.header or self._line == '' or self._line[0] == "#":
            self.__feature = None
            self.header = True
            return
        elif len(self._fields) != 9:
            raise InvalidParsingFormat("Invalid GF line: {}".format(self._line))

        self.chrom, self.source = self._fields[0:2]
        try:
            self.start, self.end = tuple(int(i) for i in self._fields[3:5])
        except (ValueError, TypeError):
            error = "Invalid start and end values: {}\n".format(" ".join(self._fields[3:5]))
            error += "Line: {}".format(self._line)
            raise ValueError(error)

        self.score = self._fields[5]
        self.strand = self._fields[6]
        self.phase = self._fields[7]

        self._attr = self._fields[8]
        self._parse_attributes()
        self.feature = self._fields[2]
        self.__is_exon, self.__is_gene, self.__is_cds = None, None, None
        [intern(_) for _ in (str(self.chrom), str(self.source), str(self.feature))]

    def __str__(self):
        if not self.feature:
            return self._line.rstrip()
        score, strand, phase = self.__format_middle()
        attributes = self._format_attributes()
        if self.source is None:
            self.source = "Mikado"
        line = [self.chrom, self.source, self.feature,
                self.start, self.end, score,
                strand, phase, attributes]
        return "\t".join([str(_) for _ in line])

    def __len__(self):
        if self.header is False and all([_ is not None for _ in [self.end, self.start]]):
            return self.end - self.start + 1
        else:
            return 0

    def __getstate__(self):
        state = {
            "chrom": self.chrom,
            "source": self.source,
            "feature": self.feature,
            "start": self.start,
            "end": self.end,
            "score": self.score,
            "strand": self.strand,
            "phase": self.phase,
            "attributes": self.attributes,
            "is_exon": self.is_exon,
            "is_cds": self.is_cds,
            "is_gene": self.is_gene,
            "transcript": self._transcript,
            "gene": self.__gene,
            "header": self.header,
            "_line": self._line,
            "_fields": self._fields
        }
        return state

    def __setstate__(self, state):
        self._transcript = state.pop("transcript")
        self.__gene = state.pop("gene")
        self.__is_gene = state.pop("is_gene")
        self.__is_cds = state.pop("is_cds")
        self.__is_exon = state.pop("is_exon")
        self.__dict__.update(state)

    def __hash__(self):
        state = self.__getstate__()
        state["_fields"] = tuple(state["_fields"])
        state["attributes"] = tuple(state["attributes"].items())
        return hash(frozenset(state.items()))

    def as_dict(self):
        return self.__getstate__()

    def load_dict(self, state):
        self.__setstate__(state)

    @classmethod
    def string_from_dict(cls, data, **kwargs):

        line = [data["chrom"],
                data["source"] if data["source"] else "Mikado",
                data["feature"],
                data["start"],
                data["end"],
                data["score"] if data["score"] is not None else ".",
                data["strand"] if data["strand"] is not None else ".",
                data["phase"] if data["phase"] is not None else "."
                ]

        line = [str(item) for item in line]
        attrs = cls._format_attributes_dict(data["attributes"], **kwargs)
        return "\t".join(line + [attrs])

    @abc.abstractmethod
    def _parse_attributes(self):
        """
        Abstract method. It is used by the children class to serialise the data
         inside the ninth field, which is one of the biggest differences between
         GFFs and GTFs.
        :return:
        """
        raise NotImplementedError("This is only an abstract method!")

    @staticmethod
    def _attribute_definition(val):
        return _attribute_definition(val)

    @abc.abstractmethod
    def _format_attributes(self):
        """ Abstract method. Each children class should implement its own version,
        which allow to prepare the ninth field for printing.
        :return:
        """
        raise NotImplementedError("This is only an abstract method!")

    @staticmethod
    @abc.abstractmethod
    def _format_attributes_dict(attributes, **kwargs):
        """Method to define how to get the attribute string."""

    def __format_middle(self):
        """
        Private method to format the middle fields (score, strand, phase)
        for printing.
        :return: score, strand, phase
        :rtype: str, str, str
        """

        score = self.score
        strand = self.strand
        phase = self.phase

        if score is None:
            score = "."
        else:
            score = int(score)
        if strand is None:
            strand = '.'
        else:
            strand = strand
        if phase is None:
            phase = "."
        return score, strand, phase

    def copy(self):
        """
        Wrapper around the copy.copy function.
        """
        return copy.copy(self)

    @property
    def feature(self):
        try:
            return self.__feature
        except AttributeError:
            self.__feature = self.__dict__["feature"]
            return self.__feature

    @feature.setter
    def feature(self, feature):
        self.__feature = feature
        self.__is_exon, self.__is_cds, self._is_gene = None, None, None

    # Instance properties
    @property
    def strand(self):

        """
        Strand attribute. One of None, +, -
        :rtype str | None
        """

        return self.__strand

    @strand.setter
    def strand(self, strand):
        """
        Strand setter. It verifies that strand is a valid value (None, +, -)
        :param strand: new strand value
        :type strand: str | None
        """

        if strand in ("+", "-"):
            self.__strand = strand
        elif strand in (None, ".", "?"):
            self.__strand = None
        else:
            raise ValueError("Invalid value for strand: {0}".format(strand))

    @property
    def score(self):
        """
        Score value. Either None or float.
        :rtype float | None
        """
        try:
            return self.__score
        except AttributeError:
            self.__score = self.__dict__["score"]
            return self.__score

    @score.setter
    def score(self, *args):
        """
        Score setter. It verifies that the score is a valid value (None or float).
        :param args: List of arguments. Only the first is used.
        :type args: list | None | float
        """

        score = args[0]
        if score == ".":
            score = None
        elif score is not None:
            try:
                score = float(args[0])
            except ValueError:
                score = None
            except SystemError:
                raise ValueError("Invalid score value: {}".format(args[0]))
        if score is not None and not isinstance(score, float):
            raise TypeError(score)
        self.__score = score

    @property
    def phase(self):
        """
        Property. Stores the phase of the feature.
        Valid values: None, 0, 1 ,2
        :return:
        """

        try:
            return self.__phase
        except AttributeError:
            self.__phase = self.__dict__["phase"]
            return self.__phase

    @phase.setter
    def phase(self, value):
        """
        Setter for the phase attribute.
        :param value:
        :return:
        """

        if value is None:
            self.__phase = None
        else:
            if value in (None, '.', '?'):
                self.__phase = None
            else:
                try:
                    phase = int(value)
                except (TypeError, ValueError):
                    raise ValueError("Invalid phase {0} (type: {1})\nLine:{2}".format(value, type(value),
                                                                                      self._line))
                if phase in (-1, 0, 1, 2):
                    phase = max(phase, 0)
                    self.__phase = phase
                else:
                    raise ValueError("Invalid phase {0} (type: {1})\nLine:{2}".format(value, type(value),
                                                                                      self._line))
        self._set_frame()

    def _set_frame(self):
        if self.phase is not None:
            self.__frame = (3 - self.__phase) % 3
        else:
            self.__frame = None

    @property
    def frame(self):
        return self.__frame

    @property
    def is_gene(self):
        """
        Property. True if the feature ends with "gene" and has no parents.
        :rtype bool
        """
        if self._is_gene is None:
            self._is_gene = self._set_is_gene()
        return self._is_gene

    def _set_is_gene(self):
        if self.feature is not None:
            if self.feature.endswith("gene") and self.feature != "mRNA_TE_gene":
                return True
            elif self.id is not None and self.id.startswith("gene:"):
                # Hack for EnsEMBL
                return True
            elif self.feature in ("sublocus", "monosublocus", "monosublocusholder"):
                return True
        return False

    @property
    def is_exon(self):
        """
        Property. True if the feature is CDS/exon/UTR/start or stop codon.
        :rtype bool
        """
        if self.__is_exon is None:
            self.__is_exon = self._set_is_exon()

        return self.__is_exon

    exon_pattern = re.compile("(cds|exon$|utr|codon|^cdna_match$|^match_part$)", flags=re.IGNORECASE)

    def _set_is_exon(self):

        feature = self.feature
        return (feature is not None) and (re.search(self.exon_pattern, feature) is not None)

    @property
    def is_cds(self):
        """This property evaluates to True if the row describes a CDS/UTR segment,
        False otherwise."""
        if self.__is_cds is None:
            self.__is_cds = self._set_is_cds()
        return self.__is_cds

    cds_pattern = re.compile("(cds|utr|codon)", flags=re.IGNORECASE)

    def _set_is_cds(self):
        return self.is_exon and (re.search(self.cds_pattern, self.feature) is not None)

    @property
    @abc.abstractmethod
    def _negative_order(self):
        raise NotImplementedError("This is implemented in the children classes")

    @property
    @abc.abstractmethod
    def _positive_order(self):
        raise NotImplementedError("This is implemented in the children classes")

    def _sort_feature(self, feature):
        """
        Private method that sorts features according to the normal order in a GF file.
        :param feature:
        :return: numeric sort index
        """

        try:
            strand = self.strand
        except AttributeError:
            strand = self.__dict__["strand"]

        if strand == "-":
            order = self._negative_order
        else:
            order = self._positive_order
        assert order != []

        if feature not in order:
            return float("inf")
        else:
            return order.index(feature)

    def __lt__(self, other):

        if self.chrom != other.chrom:
            return self.chrom < other.chrom
        else:
            if self.start != other.start:
                return self.start < other.start
            elif self.end != other.end:
                return self.end < other.end
            elif self.feature != other.feature:
                return self._sort_feature(self.feature) < self._sort_feature(other.feature)
            else:
                return False

    def remove_attribute(self, attribute):
        """Method to remove attributes from the internal dictionary."""

        if attribute in self.attributes:
            del self.attributes[attribute]
        try:
            self.attribute_order.remove(attribute)
        except ValueError:
            pass
        return

    def add_attribute(self, attribute, value):

        self.attributes[attribute] = value

        if attribute not in self.attribute_order:
            self.attribute_order.append(attribute)

    def __eq__(self, other):

        if self.is_exon is True:
            return (self.chrom == other.chrom and
                    self.feature == other.feature and
                    self.start == other.start and
                    self.end == other.end)
        else:
            return (self.id == other.id and
                    self.feature == other.feature and
                    self.chrom == other.chrom and
                    self.start == other.start and
                    self.end == other.end)



cpdef tuple compare(prediction, reference, bint lenient=?, bint strict_strandedness=?, int fuzzymatch=?)
cdef str __assign_multiexonic_ccode(prediction, reference, long nucl_overlap, double stats[9], int fuzzymatch=?)
cdef str __assign_monoexonic_ccode(prediction, reference, long nucl_overlap, double stats[9])



#!/usr/bin/env python3
# coding: utf-8

""" A clone of grep -f."""

import sys
import time
import argparse
import os


def main():
    """Main script class."""

    parser = argparse.ArgumentParser(
        description='''This script is basically an efficient version of the GNU "grep -f" utility for table-like files, and functions with a similar sintax.''')
    parser.add_argument('-v', '--reverse', action='store_true', default=False,
                        help='Equivalent to the "-v" grep option')
    parser.add_argument('-s', '--separator', type=str, default=None,
                        help='The field separator. Default: consecutive whitespace(s)')
    parser.add_argument('-f', '--field', type=int, default=1, help='The field to look in the target file.')
    parser.add_argument('-q', '--quiet', default=True, action='store_false', help='No logging.')
    parser.add_argument('ids', help='The file of patterns to extract', type=str)
    parser.add_argument('target', help='The file to filter', type=argparse.FileType())
    parser.add_argument('out', nargs='?', help='The output file', type=argparse.FileType('w'), default=sys.stdout)

    args = parser.parse_args()

    start = time.time()
    if not args.quiet:
        print(time.ctime(), 'Begin loading', file=sys.stderr)

    if os.path.exists(args.ids):
        ids_d = set([line.rstrip() for line in open(args.ids).readlines()])
    else:
        ids_d = set(args.ids.split(","))
    if not args.quiet:
        print(len(ids_d), file=sys.stderr)

    if not args.quiet:
        print(time.ctime(), 'Finished loading', file=sys.stderr)
    if args.field > 0:
        args.field -= 1

    try:
        for line in args.target:
            fields = line.rstrip().split(args.separator)
            if len(fields) <= args.field:
                continue
                # print("Too few fields ({0}), you asked for field n. {1}".format(len(fields), args.field), file=sys.stderr)
                # print(fields, file=sys.stderr)
                # raise ValueError("Wrong number of fields")
            else:
                try:
                    idl = fields[args.field]
                except IndexError:
                    raise IndexError(fields)

            if args.reverse:
                if idl not in ids_d:
                    print(line, file=args.out, end='')
            else:
                if idl in ids_d:
                    print(line, file=args.out, end='')
            if not line.strip():
                print(line, file=args.out, end='')
    except BrokenPipeError:
        pass
    except KeyboardInterrupt:
        pass

    if not args.quiet:
        print(time.ctime(), 'Finished', file=sys.stderr)
    stop = time.time()
    tim = stop - start
    if not args.quiet:
        print(time.ctime(), 'The run lasted %f seconds' % tim, file=sys.stderr)

    return 0


if __name__ == '__main__':
    main()



from ..loci import Transcript, Gene
import unittest
from .. import parsers
from pkg_resources import resource_filename


class GeneTester(unittest.TestCase):

    def test_null_creation(self):

        null_gene = Gene(None)
        self.assertIsNone(null_gene.chrom)
        self.assertIsNone(null_gene.strand)
        self.assertIsNone(null_gene.source)
        self.assertIsNone(null_gene.start)
        self.assertEqual(null_gene.feature, "gene")

    def test_creation_from_transcript(self):

        t = Transcript()
        t.chrom = "Chr1"
        t.start = 100
        t.end = 200
        t.strand = "+"
        t.id = "test"
        t.parent = "parent"
        gene = Gene(t)
        self.assertIn(t.id, gene, gene.keys())
        self.assertIn(t, gene, gene.keys())
        self.assertEqual(t.chrom, gene.chrom)
        self.assertEqual(t.start, gene.start)
        self.assertEqual(t.end, gene.end)
        self.assertEqual(t.strand, gene.strand)
        self.assertIs(t, gene[t.id])

    def test_pesky_gtf(self):
        fname = resource_filename("Mikado.tests", "gtf_check.gtf.gz")
        with parsers.GTF.GTF(fname) as gtf:
            gene = None
            for gline in gtf:
                self.assertFalse(gline.header)
                self.assertEqual(gline.gene, "ENSG00000142621")
                if gline.is_gene:
                    gene = Gene(gline)
                elif gline.is_transcript:
                    gene.add(Transcript(gline))
                else:
                    gene.add_exon(gline)

        gene.finalize()
        self.assertTrue(gene.is_coding)
        self.assertTrue(sorted(list(gene.transcripts.keys())),
                    sorted("""ENST00000314668
ENST00000358897
ENST00000375995
ENST00000375996
ENST00000375997
ENST00000375998
ENST00000433640
ENST00000444385
ENST00000459961
ENST00000462113
ENST00000471347
ENST00000472086
ENST00000472131
ENST00000472449
ENST00000477846
ENST00000481324
ENST00000483120
ENST00000483694
ENST00000495195
ENST00000524761
ENST00000529606
ENST00000532408
""".split("\n")))


if __name__ == '__main__':
    unittest.main()



"""
This module contains basic utilities for the suite, like e.g. database connection
and log creation.
"""


import os
import functools
import re
from typing import Union

import tempfile

from . import dbutils
from . import log_utils
import collections
import gzip
from itertools import zip_longest
from .overlap import overlap
from . import intervaltree
from .f1 import calc_f1
from .intervaltree import Interval, IntervalTree, IntervalNode, distance
import sys


__author__ = 'Luca Venturini'


# Diamond default: qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore
# BLASTX default: qaccver saccver pident length mismatch gapopen qstart qend sstart send evalue bitscore
from ..exceptions import InvalidConfiguration

blast_keys = "qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore ppos btop".split()


def comma_split(string):
    """Small utility to split a string based on comma. Useful for parsers."""

    return string.split(",")


def path_join(output_dir, output_file):

    """Small utility to join together a directory path and
    an output file, checking first that the output file is not
    an absolute path.

    :param output_dir: the output directory
    :param output_file: the output file
    """

    if os.path.isabs(output_file):
        return output_file
    else:
        return os.path.join(output_dir,
                            output_file)


def memoize(obj):

    """
    Function to memorize the results of functions/properties in memory for fast access.
    Source: https://wiki.python.org/moin/PythonDecoratorLibrary#Memoize

    :param obj: the object/function to memoize
    :return:
    """

    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]
    return memoizer


def merge_partial(filenames, handle, logger=None, gzipped=False):

    """This function merges the partial files created by the multiprocessing into a single
    sorted file.

    :param filenames: the filenames to merge into the handle
    :type filenames: list[str]

    :param handle: the handle to use for printing
    :type handle: io.TextIOWrapper

    :param logger: logger to be used for the merging

    """

    if logger is None:
        logger = log_utils.create_null_logger("merger")
    if len(filenames) == 0:
        logger.warning("Nothing to merge. Exiting")

    logger.debug("Starting to merge %d files (root: %s)",
                 len(filenames), "-".join(filenames[0].split("-")[:-1]))

    current_lines = collections.defaultdict(list)

    try:
        if gzipped is False:
            fnames = [open(_) for _ in filenames if os.stat(_).st_size > 0]
        else:
            fnames = [gzip.open(_, "rt") for _ in filenames if os.stat(_).st_size > 0]
    except FileNotFoundError as exc:
        raise FileNotFoundError((filenames, os.listdir(os.path.dirname(filenames[0]))))

    if len(fnames) == 0:
        logger.warning("All the files to merge (root %s) are empty. Exiting.",
                       "-".join(filenames[0].split("-")[:-1]))
        [_.close() for _ in fnames]
        return 0

    for lines in zip_longest(*fnames):
        for line in iter(_ for _ in lines if _ is not None):
            _ = line.split("/")
            index = int(_[0])
            current_lines[index].append("/".join(_[1:]))

    if len(current_lines) == 0:
        logger.exception("Nothing found to merge  for root %s. ERROR!.",
                         "-".join(filenames[0].split("-")[:-1]))
        [_.close() for _ in fnames]
        [os.remove(_) for _ in filenames]

        raise IndexError

    total = max(current_lines.keys())
    logger.debug("Merging %d lines into %s", total, handle.name)
    [_.close() for _ in fnames]
    [os.remove(_) for _ in filenames]

    for index in sorted(current_lines.keys()):
        for line in current_lines[index]:
            print(line, file=handle, end="")
        del current_lines[index]
    logger.debug("Merged %d lines into %s", total, handle.name)
    handle.flush()
    return total


def grouper(iterable, n):
    """
    Function to chunk an iterable into slices of at most n elements.
    :param iterable:
    :param n:
    :return:
    """

    temp = []
    for val in iterable:
        temp.append(val)
        if len(temp) >= n:
            yield temp
            temp = []
    if temp:
        yield temp


def merge_dictionaries(dict_a, dict_b, path=None):
    """Recursive function to merge two dictionaries.

    :param dict_a: first dictionary
    :type dict_a: dict

    :param dict_b: second dictionary
    :type dict_b: dict

    :param path: list to be updated during recursion to indicate
                 that we are in a secondary node
    :type path: list(str)

    Source: http://stackoverflow.com/questions/7204805/dictionaries-of-dictionaries-merge
    """

    if path is None:
        path = []
    for key in dict_b:
        if key in dict_a and isinstance(dict_a[key], dict) and isinstance(dict_b[key], dict):
            merge_dictionaries(
                dict_a[key],
                dict_b[key], path + [str(key)])
        else:
            dict_a[key] = dict_b[key]
    return dict_a


def merge_ranges(ranges):
    """
    Merge overlapping and adjacent ranges and yield the merged ranges
    in order. The argument must be an iterable of pairs (start, stop).

    >>> list(merge_ranges([(5,7), (3,5), (-1,3)]))
    [(-1, 7)]
    >>> list(merge_ranges([(5,6), (3,4), (1,2)]))
    [(1, 2), (3, 4), (5, 6)]
    >>> list(merge_ranges([]))
    []
    """
    ranges = iter(sorted(ranges))
    try:
        current_start, current_stop = next(ranges)
    except StopIteration:
        return
    for start, stop in ranges:
        if start > current_stop:
            # Gap between segments: output current segment and start a new one.
            yield current_start, current_stop
            current_start, current_stop = start, stop
        else:
            # Segments adjacent or overlapping: merge.
            current_stop = max(current_stop, stop)
    yield current_start, current_stop


_reg_pat = re.compile(r"^([^:]*):(\d*)(?:-|\.\.)(\d*)$")


def to_region(string: Union[str, bytes]) -> [str, int, int]:

    """
    Snippet to convert from Apollo-style region to a tuple of chrom, start, end
    :param string:
    :return:
    """

    if not isinstance(string, (str, bytes)):
        raise ValueError("Invalid region: {} (type {})".format(string, type(string)))
    elif isinstance(string, bytes):
        string = string.decode()
    string = string.strip()
    try:
        chrom, start, end = _reg_pat.search(string).groups()
        start, end = int(start), int(end)
    except (ValueError, AttributeError, TypeError):
        raise ValueError("Invalid string specified: {}".format(string))
    if end < start:
        raise ValueError("Start greater than end: {0}\t{1}".format(start, end))
    return chrom, start, end


def percentage(value):
    value = float(value)
    if value < 0:
        raise ValueError("Negative numbers are not allowed")
    elif value > 100:
        raise ValueError("Only numbers between 0 and 100 are allowed")
    while 1 < value <= 100:
        value /= 100
    return value


def default_for_serialisation(obj):
    if isinstance(obj, set):
        return tuple(obj)
    elif obj == float("inf"):
        return sys.maxsize


def to_bool(param: Union[str, bool, int, float]):
    """Function to convert a items to booleans."""

    if isinstance(param, bool):
        return param
    elif isinstance(param, (int, float)):
        if param == 1:
            return True
        elif param == 0:
            return False
    elif isinstance(param, (str, bytes)):
        if isinstance(param, bytes):
            param = param.decode()
        lparam = param.lower()
        if lparam == 'true' or lparam == "1":
            return True
        elif lparam == 'false' or lparam == "0":
            return False

    raise ValueError(f"Invalid boolean parameter: {param}")


def create_shm_handle():
    shm_available = os.path.exists("/dev/shm") and os.access("/dev/shm", os.W_OK)
    if shm_available is False:
        return None
    try:
        shm_handle = tempfile.NamedTemporaryFile(suffix=".db", dir="/dev/shm/")
        return shm_handle
    except PermissionError:
        return None

    # if self.configuration.pick.run_options.shm is True:
    #
    #     if shm_available is False:
    #         self.main_logger.info("Mikado was asked to copy the database into /dev/shm, but it is either \
    # not available or not writable for this user. Leaving the DB where it is.")
    #         return
    #     self.main_logger.info("Copying Mikado database into a SHM db")
    #     assert self.configuration.db_settings.dbtype == "sqlite"
    #     # Create temporary file
    #     self.shm_db = tempfile.NamedTemporaryFile(suffix=".db", dir="/dev/shm/")
    #     self.main_logger.debug("Copying {0} into {1}".format(
    #         self.configuration.db_settings.db,
    #         self.shm_db.name))
    #     try:
    #         shutil.copy2(self.configuration.db_settings.db, self.shm_db.name)
    #         self.configuration.db_settings.db = self.shm_db.name
    #     except PermissionError:
    #         self.main_logger.warning(
    #             """Permission to write on /dev/shm denied.
    #             Back to using the DB on disk.""")
    #         self.configuration.pick.run_options.shm = False
    #     self.main_logger.info("DB copied into memory")


from dataclasses import field
from marshmallow import validate
from marshmallow_dataclass import dataclass, Optional, List
from .configuration import MikadoConfiguration


@dataclass
class ShortReads:
    r1: List[str] = field(default_factory=lambda: [], metadata={
        "metadata": {"description": "Array of left read files."},
        "required": True
    })
    r2: List[str] = field(default_factory=lambda: [], metadata={
        "metadata": {
            "description": "Array of right read files. It must be of the same length of r1; if one or more of the \
samples are single-end reads, add an empty string."},
        "required": True
    })
    samples: List[str] = field(default_factory=lambda: [], metadata={
        "metadata": {"description": "Array of the sample names. It must be of the same length of r1."},
        "required": True
    })
    strandedness: List[str] = field(default_factory=lambda: [], metadata={
        "metadata": {
            "description": "Array of strand-specificity of the samples. It must be of the same length of r1. \
Valid values: fr-firststrand, fr-secondstrand, fr-unstranded."},
        "required": True
    })


@dataclass
class LongReads:
    files: List[str] = field(default_factory=lambda: [])
    samples: List[str] = field(default_factory=lambda: [])
    strandedness: List[str] = field(default_factory=lambda: [])
    skip_split: bool = field(default=True)


@dataclass
class OrfCalling:
    min_protein_len: int = field(default=30, metadata={
        "metadata": {"description": "minimum length of called proteins (in AAs). Default: 30 (90 nts)"},
        "validate": validate.Range(min=0)
    })
    execute: bool = field(default=True, metadata={
        "metadata": {"description": "boolean flag. Default: true, ie execute the ORF calling."}
    })


@dataclass
class BlastX:
    prot_db: Optional[List[str]] = field(default_factory=lambda: [], metadata={
        "metadata": {
            "description": "FASTA file(s) to be used as database for the homology search. Multiple files can be provided and they will be merged into a single database before running."},
    })
    evalue: float = field(default=1e-7, metadata={
        "metadata": {"description": "Maximum e-value. Default 1e-7"},
        "validate": validate.Range(min=0)
    })
    max_target_seqs: int = field(default=10, metadata={
        "metadata": {"description": "Maximum number of targets that the homology will report. Default: 10"},
        "validate": validate.Range(min=1)
    })
    chunks: int = field(default=10, metadata={
        "metadata": {
            "description": "Number of chunks to divide the search into. Must be equal or greater than the number of processes."
        },
        "validate": validate.Range(min=1)
    })


@dataclass
class AlignMethods:
    tophat: List[str] = field(default_factory=lambda: [])
    hisat: List[str] = field(default_factory=lambda: [])
    star: List[str] = field(default_factory=lambda: [])
    gsnap: List[str] = field(default_factory=lambda: [])


@dataclass
class LongReadAlign:
    star: List[str] = field(default_factory=lambda: [])
    gmap: List[str] = field(default_factory=lambda: [])


@dataclass
class AsmMethods:
    stringtie: List[str] = field(default_factory=lambda: [])
    cufflinks: List[str] = field(default_factory=lambda: [])
    class2: List[str] = field(default_factory=lambda: [])
    scallop: List[str] = field(default_factory=lambda: [])
    trinity: List[str] = field(default_factory=lambda: [])


@dataclass
class ProgramLoader:
    tophat: str = field(default='')
    gmap: str = field(default='')
    star: str = field(default='')
    hisat: str = field(default='')
    samtools: str = field(default='')
    cufflinks: str = field(default='')
    scallop: str = field(default='')
    trinity: str = field(default='')
    stringtie: str = field(default='')
    class2: str = field(default='')
    transdecoder: str = field(default='')
    prodigal: str = field(default='')
    portcullis: str = field(default='')
    mikado: str = field(default='')
    diamond: str = field(default='')
    blast: str = field(default='')


@dataclass
class Portcullis:
    do: bool = field(default=True, metadata={"required": True})
    canonical_juncs: str = field(default="C,S", metadata={"required": True})


@dataclass
class AlnIndex:
    star: str = field(default="")


@dataclass
class DaijinMikadoConfiguration:
    use_diamond: bool = field(default=True, metadata={
        "metadata": {"description": "use DIAMOND instead of NCBI BLASTX. Default: true"},
    })
    use_prodigal: bool = field(default=True, metadata={
        "metadata": {"description": "Use Prodigal instead of TransDecoder for ORF calling. Default: true"}
    })
    modes: List[str] = field(default_factory=lambda: ["stringent"], metadata={
        "metadata": {
            "description": "which mode(s) to run Mikado into. Default: permissive (split multiple ORF models \
unless there is strong BLAST evidence against the decision)."},
        "validate": validate.ContainsOnly(["nosplit", "split", "lenient", "stringent", "permissive"]),
        "required": True
    })


@dataclass
class TGGConfiguration:
    max_mem: int = field(default=6000, metadata={
        "metadata": {"description": "Maximum memory to be used for the assembly. Default: 6000Mb"},
        "validate": validate.Range(min=1000)
    })
    npaths: int = field(default=0, metadata={
        "metadata": {
            "description": "Number of alignments per sequence, using GMAP. Default: 0 (one alignment per sequence, exclude chimeric)."}
    })
    identity: float = field(default=0.95, metadata={
        "metadata": {"description": "minimum identity for any alignment. Default: 95%"},
        "validate": validate.Range(min=0.0, max=1.0)
    })
    coverage: float = field(default=0.70, metadata={
        "metadata": {"description": "minimum coverage for any alignment. Default: 70%"},
        "validate": validate.Range(min=0.0, max=1.0)
    })


@dataclass
class DaijinConfiguration(MikadoConfiguration):
    """
    Configuration properties for Daijin. This is an extended Mikado configuration file and can be given directly to
Mikado itself.
    """

    load_cmd: ProgramLoader = field(default_factory=ProgramLoader, metadata={
                "metadata": {
            "description": "Commands to use to load/select the versions of the programs to use. Leave an empty string if no loading is necessary."},
    })
    portcullis: Portcullis = field(default_factory=Portcullis, metadata={
                "metadata": {"description": "Options related to portcullis"},
    })
    aln_index: AlnIndex = field(default_factory=AlnIndex, metadata={
                "metadata": {"description": "Options related to indexing."},
    })
    long_reads: LongReads = field(default_factory=LongReads, metadata={
                "metadata": {"description": "Parameters related to long reads to use for the assemblies."},
    })
    short_reads: ShortReads = field(default_factory=ShortReads, metadata={
                "metadata": {"description": "Parameters related to the reads to use for the assemblies."},
    })
    name: str = field(default="Daijin", metadata={
        "metadata": {"description": "Name to be used for the project"},
        "required": True,
        "validate": validate.Length(min=1)
    })
    out_dir: str = field(default="Daijin", metadata={
        "metadata": {"description": "Output directory for the project"},
        "required": True,
        "validate": validate.Length(min=1)
    })
    scheduler: Optional[str] = field(default=None, metadata={
        "metadata": {
            "description": "Scheduler to be used for the project. Set to null if you plan to use DRMAA or are using a local machine."
        },
        "validate": validate.OneOf(["SLURM", "LSF", "local", "PBS", ""])
    })
    align_methods: AlignMethods = field(default_factory=AlignMethods, metadata={
                "metadata": {
            "description": "Aligners to use. Each aligner can be invoked multiple times: the per-aligner list includes the extra command line arguments to be passed to the program"},
    })
    long_read_align_methods: LongReadAlign = field(default_factory=LongReadAlign, metadata={
                "metadata": {
            "description": "Aligners for long reads to use. Each aligner can be invoked multiple times: the per-aligner list includes the extra command line arguments to be passed to the program"},
    })
    asm_methods: AsmMethods = field(default_factory=AsmMethods, metadata={
                "metadata": {
            "description": "Short-read assemblers to use. Each assembler can be invoked multiple times: the per-aligner list includes the extra command line arguments to be passed to the program"},
    })
    orf_calling: OrfCalling = field(default_factory=OrfCalling, metadata={
                "metadata": {"description": "Parameters related to the ORF calling:"},
    })
    blastx: BlastX = field(default_factory=BlastX, metadata={
                "metadata": {"description": "Parameters related to how DIAMOND/BLASTX will be run"}
    })
    mikado: DaijinMikadoConfiguration = field(default_factory=DaijinMikadoConfiguration, metadata={
        "metadata": {"description": "Parameters related to the Mikado execution."},
            })
    tgg: TGGConfiguration = field(default_factory=TGGConfiguration, metadata={
        "metadata": {"description": "Options related to genome-guided Trinity."},
            })



import cython

@cython.profile(True)
cpdef long overlap(first, second, long flank=0, bint positive=0):

    """This function quickly computes the overlap between two
    ranges, with an optional flank."""

    cdef long start, end, ostart, oend

    if hasattr(first, "start"):
        start, end = first.start, first.end
    else:
        start, end = first[:2]
    if hasattr(second, "start"):
        ostart, oend = second.start, second.end
    else:
        ostart, oend = second[:2]

    return c_overlap(start, end, ostart, oend, flank=flank, positive=positive)


@cython.profile(True)
cdef long c_overlap(long start, long end, long ostart, long oend, long flank, bint positive):
    if start > end:
        start, end = end, start
    if ostart > oend:
        ostart, oend = oend, ostart

    cdef long right
    cdef long left

    cdef long left_one = start - flank
    cdef long left_two = ostart - flank

    if left_one > left_two:
        left = left_one
    else:
        left = left_two

    cdef long right_one = end + flank
    cdef long right_two = oend + flank

    if right_one < right_two:
        right = right_one
    else:
        right = right_two

    cdef long result = right - left
    if positive == 1 and result < 0:
        return 0
    else:
        return result



# coding: utf-8

"""
This module is necessary to serialise the junction information
provided in BED12 format by programs such as TopHat or portcullis.
Please convert any other (custom) format into BED12
before loading.
"""

import io
import os
from collections import namedtuple

from sqlalchemy import Column
from sqlalchemy import String
from sqlalchemy import Integer
from sqlalchemy import ForeignKey
from sqlalchemy import CHAR
from sqlalchemy import Index
from sqlalchemy import Float
from sqlalchemy import select
from sqlalchemy.orm import relationship, column_property
from sqlalchemy.orm.session import Session
from sqlalchemy.ext.hybrid import hybrid_method
from ..utilities.dbutils import DBBASE, connect
import sqlalchemy as sqla
from ..parsers import bed12
from ..utilities.log_utils import check_logger, create_default_logger
import pyfaidx
from copy import deepcopy as copy
from ..configuration.configuration import MikadoConfiguration
from ..configuration.daijin_configuration import DaijinConfiguration


# pylint: disable=too-few-public-methods
class Chrom(DBBASE):
    """
    Simple serialization class for chromosomes.
    """

    __tablename__ = "chrom"
    __table_args__ = {"extend_existing": True}

    chrom_id = Column(Integer, primary_key=True)
    name = Column(String(200), unique=True)
    length = Column(Integer, nullable=True)

    def __init__(self, name, length=None):
        self.name = name
        if length is not None:
            assert isinstance(length, int)
        self.length = length


# A serialisation class must have a ton of attributes ...
# pylint: disable=too-many-instance-attributes
class Junction(DBBASE):
    """
    Class that describes the junction table in the database.
    """

    __tablename__ = "junctions"

    # pylint: disable=invalid-name
    id = Column(Integer, primary_key=True)
    # pylint: enable=invalid-name
    chrom_id = Column(Integer, ForeignKey(Chrom.chrom_id), unique=False)
    start = Column(Integer, nullable=False)
    end = Column(Integer, nullable=False)
    name = Column(String(200))
    strand = Column(CHAR)
    junction_start = Column(Integer, nullable=False)
    junction_end = Column(Integer, nullable=False)
    score = Column(Float)
    __table_args__ = (Index(
        "junction_index", "chrom_id", "junction_start", "junction_end"), {"extend_existing": True})

    chrom_object = relationship(Chrom, uselist=False)
    chrom = column_property(select([Chrom.name]).where(chrom_id == Chrom.chrom_id).scalar_subquery())

    def __init__(self, junction_start, junction_end, name, strand, score, chrom_id):
        """
        Serialization initializer.
        :param bed12_object: a BED12-like class instance.
        :type bed12_object: Mikado.parsers.bed12.BED12

        :param chrom_id: the numerical foreign key for the Chrom table.
        :type chrom_id: int
        """

        # if not isinstance(bed12_object, bed12.BED12):
        #     raise TypeError("Invalid data type!")
        self.chrom_id = chrom_id
        self.start = junction_start
        self.end = junction_end
        self.junction_start = junction_start
        self.junction_end = junction_end
        self.name = name
        self.strand = strand
        self.score = score

    def __eq__(self, other):
        return (self.chrom_id, self.start, self.end, self.strand) == (other.chrom_id, other.start, other.end, other.strand)

    def __hash__(self):
        return hash((self.chrom_id, self.start, self.end, self.strand))

    def __str__(self):
        return "{chrom}\t{start}\t{end}\t{strand}".format(
            chrom=self.chrom,
            start=self.start,
            end=self.end,
            strand=self.strand
        )

    def __repr__(self):
        return "{chrom}\t{start}\t{end}\t{strand}".format(
            chrom=self.chrom,
            start=self.start,
            end=self.end,
            strand=self.strand
        )

    @hybrid_method
    def is_equal(self, chrom, start, end, strand):
        """
        Function to verify whether a set of coordinates is equal to those in the DB.

        :param chrom: chromosome
        :type chrom: str

        :param start: start position (1-offset)
        :type start: int

        :param end: end position (1-offset)
        :type end: int

        :param strand: strand of the junction
        :type strand: str
        """
        # print(self.chrom_object.chrom_id)

        return ((self.chrom == chrom) and (self.start == start) and
                (self.end == end) and (self.strand == strand))
# pylint: enable=too-many-instance-attributes


class JunctionSerializer:

    """
    This class is used to serialize a junction BED12 file into an SQL database.
    """

    def __init__(self, handle,
                 configuration=None,
                 logger=None):

        """
        :param handle: the file to be serialized.
        :type handle: str | io.IOBase | io.TextIOWrapper

        :param configuration: Optional configuration dictionary with db connection parameters.
        :type configuration: (MikadoConfiguration|DaijinConfiguration)
        """

        self.bed12_parser = None
        self.fai = None

        if logger is not None:
            logger = check_logger(logger)
            self.logger = logger
        else:
            self.logger = create_default_logger("junction")

        if handle is None:
            self.logger.warning("No input file specified. Exiting.")
            self.close()
            return

        self.bed12_parser = bed12.Bed12Parser(handle)
        self.engine = connect(configuration, logger=logger)
        self.db_settings = copy(configuration.db_settings)

        session = Session(bind=self.engine, autocommit=False, autoflush=False, expire_on_commit=False)
        inspector = sqla.inspect(self.engine)
        if Junction.__tablename__ not in inspector.get_table_names():
            DBBASE.metadata.create_all(self.engine)  # @UndefinedVariable

        self.session = session
        if configuration is not None:
            self.maxobjects = configuration.serialise.max_objects
        else:
            self.maxobjects = 10000

        if not configuration.reference.genome_fai:
            _ = pyfaidx.Fasta(configuration.reference.genome)
            self.fai = _.faidx.indexname
            _.close()
        else:
            self.fai = configuration.reference.genome_fai

        if isinstance(self.fai, str):
            assert os.path.exists(self.fai), "File {fai} does not exist!".format(fai=self.fai)
            self.fai = open(self.fai)
        else:
            if self.fai is not None:
                if not isinstance(self.fai, io.TextIOWrapper):
                    msg = "The FAI index is not a valid file handle. Type: {}".format(type(self.fai))
                    logger.critical(msg)
                    raise TypeError(msg)

    def serialize(self):
        """
        Workhorse of the class. It parses the input file and loads it into the database.
        """

        sequences = dict()
        if self.bed12_parser is None:
            self.logger.warning("No input file specified. Exiting.")
            return

        self.logger.debug("Starting to serialise junctions.")
        if self.fai is not None:
            self.session.begin(subtransactions=True)
            for line in self.fai:
                name, length = line.rstrip().split()[:2]
                if self.session.query(Chrom).filter(Chrom.name == name).all():
                    continue
                try:
                    current_chrom = Chrom(name, length=int(length))
                except ValueError:
                    raise ValueError(line)
                self.session.add(current_chrom)
                sequences[current_chrom.name] = current_chrom.chrom_id
            self.session.commit()
            for query in self.session.query(Chrom):
                sequences[query.name] = query.chrom_id
            self.fai.close()

        self.logger.debug("Serialised sequences.")
        objects = []

        for counter, row in enumerate(self.bed12_parser, 1):
            if row.header is True:
                continue
            if row.chrom in sequences:
                current_chrom = sequences[row.chrom]
            else:
                current_chrom = Chrom(row.chrom)
                self.session.begin(subtransactions=True)
                self.session.add(current_chrom)
                self.session.commit()
                sequences[current_chrom.name] = current_chrom.chrom_id
                current_chrom = current_chrom.chrom_id

            self.generate_introns(current_chrom, objects, row)

            if len(objects) >= self.maxobjects:
                self.logger.debug("Serializing %d objects", len(objects))
                self.session.begin(subtransactions=True)
                self.session.bulk_save_objects(set(objects))
                self.session.commit()
                objects = []

        self.session.bulk_save_objects(set(objects))
        self.logger.debug("Serialised %s junctions into %s.", counter, self.db_settings)
        self.session.commit()
        self.close()

    @staticmethod
    def generate_introns(current_chrom, objects, row):
        # Now generate as many junctions as block start/size are present in this bed row
        # These are the exons
        Exon = namedtuple('Exon', 'start end')
        exons = [Exon(int(block[0]), int(block[1])) for block in row.blocks]
        introns = [(e0.end + 1, e1.start - 1) for e0, e1 in zip(exons, exons[1:])]
        for intron in introns:
            current_junction = Junction(intron[0], intron[1], row.name, row.strand, row.score, current_chrom)
            objects.append(current_junction)

    def __call__(self):
        """
        Alias for serialize()
        """

        self.serialize()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """
        Closing method for with. It will close the handles, if they are defined.
        :return:
        """

        if self.bed12_parser is not None:
            self.bed12_parser.close()
        if self.fai is not None:
            self.fai.close()

# pylint: enable=too-few-public-methods



#!/usr/bin/env python3

"""
This module contains generic-purpose utilities to deal with BLAST XML files.
"""

import os
import subprocess
import gzip
import io
from ..exceptions import InvalidParsingFormat
from .parser import HeaderError
from operator import itemgetter
from Bio.SearchIO import parse as bio_parser
import functools
import xml.etree.ElementTree
import numpy as np
xparser = functools.partial(bio_parser, format="blast-xml")


__author__ = 'Luca Venturini'


class BlastOpener:

    def __init__(self, filename, create_parser=True):

        self.__filename = filename
        self.__handle = None
        self.__closed = False
        self.__opened = False
        self.__parser_created = False
        self.__create_parser = create_parser
        self.__enter__()

    def __create_handle(self):

        if self.__closed is True:
            raise ValueError('I/O operation on closed file.')
        if self.__opened is True:
            return self

        if isinstance(self.__filename, (gzip.GzipFile, io.TextIOWrapper)):
            self.__handle = self.__filename
            self.__filename = self.__handle.name
        elif not isinstance(self.__filename, str) or not os.path.exists(self.__filename):
            raise OSError("Non-existent file: {0}".format(self.__filename))
        else:
            if self.__filename.endswith(".gz"):
                if self.__filename.endswith(".xml.gz"):
                    self.__handle = gzip.open(self.__filename, "rt")
                elif self.__filename.endswith(".asn.gz"):
                    # I cannot seem to make it work with gzip.open
                    zcat = subprocess.Popen(["gzip", "-dc", self.__filename], shell=False,
                                            stdout=subprocess.PIPE)
                    blast_formatter = subprocess.Popen(
                        ['blast_formatter', '-outfmt', '5', '-archive', '-'],
                        shell=False, stdin=zcat.stdout, stdout=subprocess.PIPE)
                    self.__handle = io.TextIOWrapper(blast_formatter.stdout, encoding="UTF-8")
                else:
                    raise InvalidParsingFormat("Unrecognized file format for {}".format(self.__filename))
            elif self.__filename.endswith(".xml"):
                self.__handle = open(self.__filename)
            elif self.__filename.endswith(".asn"):
                blast_formatter = subprocess.Popen(
                    ['blast_formatter', '-outfmt', '5', '-archive', self.__filename],
                    shell=False, stdout=subprocess.PIPE)
                self.__handle = io.TextIOWrapper(blast_formatter.stdout, encoding="UTF-8")
            else:
                raise InvalidParsingFormat("Unrecognized file format: {}".format(self.__filename))

        if self.__handle is None:
            raise OSError("Failed to open file {}".format(self.__filename))
        self.__opened = True

    def __enter__(self):

        try:
            self.__create_handle()
            if self.__create_parser is True and self.__parser_created is False:
                self.parser = xparser(self.__handle)
                self.__parser_created = True
            return self
        except xml.etree.ElementTree.ParseError:
            self.__closed = True

    def open(self):
        self.__enter__()

    def __exit__(self, *args):
        _ = args
        if self.__closed is False:
            self.__handle.close()
            self.__closed = True
            self.__opened = False

    def __iter__(self):
        return self

    def __next__(self):
        if self.__handle is None:
            raise ValueError("Invalid handle")
        return next(iter(self.parser))

    @property
    def handle(self):
        """Direct access to the underlying handle, for low-level access.
        This property cannot be set directly, it will be generated by the init() method of the class."""
        print(self.__handle, type(self.__handle))
        return self.__handle

    @property
    def closed(self):
        return self.__handle.closed

    @property
    def name(self):
        return self.__filename

    def close(self):
        self.__exit__()

    def read(self, *args):
        return self.__handle.read(*args)

    def sniff(self, default_header=None):

        """
        Method that either derives the default XML header for the instance (if undefined)
        or checks that the given file is compatible with it.
        :param default_header: optional default header to check for consistency
        :return: boolean (passed or not passed)
        :rtype: (bool, list, str)
        """

        if self.__opened is True:
            self.close()
            self.__closed = False
            self.__opened = False
        try:
            self.__create_handle()
        except xml.etree.ElementTree.ParseError as exc:
            return False, None, xml.etree.ElementTree.ParseError("{0} ({1})".format(exc,
                                                                                    self.__filename))

        header = []
        exc = None
        valid = True
        while True:
            try:
                line = next(self.__handle)
            except StopIteration:
                # Hack for empty files
                valid = False
                exc = HeaderError("Invalid header for {0}:\n\n{1}".format(
                    self.__filename,
                    "\n".join(header)
                ))
                break
            if "<Iteration>" in line:
                break
            line = line.rstrip()
            if not line:
                valid = False
                exc = HeaderError("Invalid header for {0}:\n\n{1}".format(
                    self.__filename,
                    "\n".join(header)
                ))
                break
            if len(header) > 10**3:
                exc = HeaderError("Abnormally long header ({0}) for {1}:\n\n{2}".format(
                    len(header),
                    self.__filename,
                    "\n".join(header)
                ))
                break
            header.append(line)
        if not any(iter(True if "BlastOutput" in x else False for x in header)):
            exc = HeaderError("Invalid header for {0}:\n\n{1}".format(
                self.__filename, "\n".join(header)))

        if default_header is not None and exc is None:
                checker = [header_line for header_line in header if
                           "BlastOutput_query" not in header_line]
                previous_header = [header_line for header_line in default_header if
                                   "BlastOutput_query" not in header_line]
                if checker != previous_header:
                    exc = HeaderError("BLAST XML header does not match for {0}".format(
                        self.__filename))
        elif exc is None:
            default_header = header

        if exc is not None:
            valid = False
        self.__handle.close()

        return valid, default_header, exc


def merge(intervals: [(int, int)], query_length=None, offset=1):
    """
    This function is used to merge together intervals, which have to be supplied as a list
    of duplexes - (start,stop). The function will then merge together overlapping tuples and
    return a list of non-overlapping tuples.
    If the list is composed by only one element, the function returns immediately.
    :param intervals: a list of integer duplexes
    :type intervals: list

    :param query_length: pre-defined length of the feature to calculate the intervals for.
    If provided, it will verify that the total sum of the ranges is within the length.
    :param offset: offset to subtract when calculating the total length. Default 1, it can be either 0 or 1.

    :returns: merged intervals, length covered



    """

    # Assume tuple of the form (start,end)
    # Create array and sort
    offset = int(offset)
    if offset not in [0, 1]:
        raise ValueError("Invalid offset - only 0 and 1 allowed: {}".format(offset))

    try:
        intervals = np.array(sorted([sorted(_) for _ in intervals], key=itemgetter(0)), dtype=int)
        if intervals.shape[1] != 2:
            raise ValueError("Invalid shape for intervals: {}".format(intervals.shape))
    except (TypeError, ValueError):
        raise TypeError("Invalid array for intervals: {}".format(intervals))
    intervals.sort()
    starts = intervals[:, 0]
    ends = np.maximum.accumulate(intervals[:, 1])
    valid = np.zeros(len(intervals) + 1, dtype=bool)
    valid[0], valid[1:-1], valid[-1] = True, starts[1:] >= ends[:-1], True
    intervals = np.vstack((starts[:][valid[:-1]], ends[:][valid[1:]])).T
    total_length_covered = int(abs(intervals[:, 1] - intervals[:, 0] + offset).sum())

    assert not (query_length and total_length_covered > query_length), \
        "Something went wrong, original length {}, total length {}".format(
            query_length, total_length_covered)

    return [tuple([int(_[0]), int(_[1])]) for _ in intervals], total_length_covered
